"""
Vector control methods for induction motor drives.

The algorithms are written based on the inverse-Γ model.

"""
from typing import Callable
from dataclasses import dataclass, field
import numpy as np

from motulator.helpers import abc2complex, Bunch
from motulator.control.common import Ctrl, SpeedCtrl, PWM


# %%
@dataclass
class InductionMotorVectorCtrlPars:
    """Vector control parameters for induction motor drives."""

    # pylint: disable=too-many-instance-attributes
    # Speed reference (in electrical rad/s)
    w_m_ref: Callable[[float], float] = field(
        repr=False, default=lambda t: (t > .2)*(2*np.pi*50))
    # Mode
    sensorless: bool = True
    # Sampling period
    T_s: float = 250e-6
    # Bandwidths
    alpha_c: float = 2*np.pi*200
    alpha_o: float = 2*np.pi*40  # Used only in the sensorless mode
    alpha_s: float = 2*np.pi*4
    # Sensored observer
    g = .2  # Used only in the sensored mode
    # Maximum values
    tau_M_max: float = 1.5*14.6
    i_s_max: float = 1.5*np.sqrt(2)*5
    # Nominal values
    psi_R_nom: float = .9
    u_dc_nom: float = 540
    # Motor parameter estimates (inverse-Γ model)
    R_s: float = 3.7
    R_R: float = 2.1
    L_sgm: float = .021
    L_M: float = .224
    n_p: int = 2
    J: float = .015


# %%
class InductionMotorVectorCtrl(Ctrl):
    """
    Vector control for an induction motor drive.

    This class interconnects the subsystems of the control system and
    provides the interface to the solver.

    Parameters
    ----------
    pars : InductionMotorVectorControlPars
        Control parameters.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        super().__init__()
        self.T_s = pars.T_s
        self.w_m_ref = pars.w_m_ref
        self.sensorless = pars.sensorless
        self.n_p = pars.n_p
        self.speed_ctrl = SpeedCtrl(pars)
        self.current_ref = CurrentRef(pars)
        self.current_ctrl = CurrentCtrl(pars)
        if self.sensorless:
            self.observer = SensorlessObserver(pars)
        else:
            self.observer = Observer(pars)
        self.pwm = PWM(pars)

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : InductionMotorDrive
            Continuous-time model of an induction motor drive for getting the
            feedback signals.

        Returns
        -------
        T_s : float
            Sampling period.
        d_abc_ref : ndarray, shape (3,)
            Duty ratio references.

        """
        # Get the speed reference
        w_m_ref = self.w_m_ref(self.t)

        # Measure the feedback signals
        i_s_abc = mdl.motor.meas_currents()  # Phase currents
        u_dc = mdl.conv.meas_dc_voltage()  # DC-bus voltage

        if not self.sensorless:
            w_m = self.n_p*mdl.mech.meas_speed()  # Rotor speed
        else:
            w_m = self.observer.w_m  # Get the estimated speed

        # Get the states
        u_s = self.pwm.realized_voltage
        psi_R = self.observer.psi_R
        theta_s = self.observer.theta_s

        # Space vector and coordinate transformation
        i_s = np.exp(-1j*theta_s)*abc2complex(i_s_abc)

        # Outputs
        tau_M_ref = self.speed_ctrl.output(w_m_ref/self.n_p, w_m/self.n_p)
        i_s_ref, tau_M_ref_lim = self.current_ref.output(tau_M_ref, psi_R)
        w_s = self.observer.output(u_s, i_s, w_m)  # w_m not used if sensorless
        u_s_ref = self.current_ctrl.output(i_s_ref, i_s)
        d_abc_ref, u_s_ref_lim = self.pwm.output(u_s_ref, u_dc, theta_s, w_s)

        # Save data
        data = Bunch(
            i_s=i_s,
            i_s_ref=i_s_ref,
            psi_R=psi_R,
            t=self.t,
            tau_M_ref_lim=tau_M_ref_lim,
            theta_s=theta_s,
            u_dc=u_dc,
            u_s=u_s,
            w_m=w_m,
            w_m_ref=w_m_ref,
            w_s=w_s,
        )
        self.save(data)

        # Update the states
        self.pwm.update(u_s_ref_lim)
        self.speed_ctrl.update(tau_M_ref_lim)
        self.current_ref.update(u_s_ref, u_dc)
        self.current_ctrl.update(u_s_ref_lim, w_s)
        self.observer.update(i_s, w_s)
        self.update_clock(self.T_s)

        return self.T_s, d_abc_ref


# %%
class CurrentRef:
    """
    Current reference calculation.

    This method includes field-weakenting operation based on the unlimited
    voltage reference feedback. The breakdown torque and current limits are
    taken into account.

    Parameters
    ----------
    pars : InductionMotorVectorCtrlPars
        Control parameters.

    Notes
    -----
    The field-weakening method and its tuning corresponds roughly to [1]_.

    References
    ----------
    .. [1] Hinkkanen, Luomi, "Braking scheme for vector-controlled induction
       motor drives equipped with diode rectifier without braking resistor,"
       IEEE Trans. Ind. Appl., 2006, https://doi.org/10.1109/TIA.2006.880852

    """

    def __init__(self, pars):
        self.T_s = pars.T_s
        self.i_s_max = pars.i_s_max
        self.L_sgm = pars.L_sgm
        self.n_p = pars.n_p
        # Local parameters for initializing the constants
        psi_R_nom = pars.psi_R_nom
        L_M = pars.L_M
        R_R = pars.R_R
        u_dc_nom = pars.u_dc_nom
        # Nominal d-axis current
        self.i_sd_nom = psi_R_nom/L_M
        # Field weakening
        self.gain_fw = 3*R_R*psi_R_nom/(pars.L_sgm*u_dc_nom)**2
        # State variable
        self.i_sd_ref = self.i_sd_nom

    def output(self, tau_M_ref, psi_R):
        """
        Compute the stator current reference.

        Parameters
        ----------
        tau_M_ref : float
            Torque reference.
        psi_R : float
            Rotor flux magnitude.

        Returns
        -------
        i_s_ref : complex
            Stator current reference.
        tau_M : float
            Limited torque reference.

        """

        def q_axis_current_limit(i_sd_ref, psi_R):
            # Priority given to the d component
            i_sq_max1 = np.sqrt(self.i_s_max**2 - i_sd_ref**2)
            # Breakdown torque limit
            i_sq_max2 = psi_R/self.L_sgm + i_sd_ref
            # q-axis current limit
            i_sq_max = np.min([i_sq_max1, i_sq_max2])
            return i_sq_max

        # q-axis current reference
        i_sq_ref = tau_M_ref/(1.5*self.n_p*psi_R) if psi_R > 0 else 0

        # Limit the current
        i_sq_max = q_axis_current_limit(self.i_sd_ref, psi_R)
        if np.abs(i_sq_ref) > i_sq_max:
            i_sq_ref = np.sign(i_sq_ref)*i_sq_max

        # Current reference
        i_s_ref = self.i_sd_ref + 1j*i_sq_ref

        # Limited torque (for the speed controller)
        tau_M_ref_lim = 1.5*self.n_p*psi_R*i_sq_ref

        return i_s_ref, tau_M_ref_lim

    def update(self, u_s_ref, u_dc):
        """
        Field-weakening based on the unlimited reference voltage.

        Parameters
        ----------
        u_s_ref : complex
            Unlimited stator voltage reference.
        u_dc : float
            DC-bus voltage.

        """
        u_s_max = u_dc/np.sqrt(3)
        self.i_sd_ref += self.T_s*self.gain_fw*(
            u_s_max**2 - np.abs(u_s_ref)**2)
        if self.i_sd_ref > self.i_sd_nom:
            self.i_sd_ref = self.i_sd_nom
        elif self.i_sd_ref < -self.i_s_max:
            self.i_sd_ref = -self.i_s_max


# %%
class CurrentCtrl:
    """
    2DOF PI current controller.

    This controller corresponds to [2]_. The continuous-time complex-vector
    design corresponding to (13) is used here. The rotor flux linkage is
    considered as a quasi-constant disturbance. This design could be
    equivalently presented as a 2DOF PI controller.

    Parameters
    ----------
    pars : InductionMotorVectorCtrlPars
        Control parameters.

    Notes
    -----
    This implementation does not take the magnetic saturation into account.

    References
    ----------
    .. [2] Awan, Saarakkala, Hinkkanen, "Flux-linkage-based current control of
       saturated synchronous motors," IEEE Trans. Ind. Appl. 2019,
       https://doi.org/10.1109/TIA.2019.2919258

    """

    def __init__(self, pars):
        self.T_s = pars.T_s
        self.L_sgm = pars.L_sgm
        self.alpha_c = pars.alpha_c
        # Integral state
        self.u_i = 0
        # Memory for the update method
        self.e = 0
        self.u_s_ref = 0

    def output(self, i_s_ref, i_s):
        """
        Compute the unlimited voltage reference.

        Parameters
        ----------
        i_s_ref : complex
            Stator current reference.
        i_s : complex
            Measured stator current.

        Returns
        -------
        u_s_ref : complex
            Unlimited voltage reference.

        """
        psi_sgm_ref = self.L_sgm*i_s_ref
        psi_sgm = self.L_sgm*i_s
        k_t = self.alpha_c
        k = 2*self.alpha_c
        self.u_s_ref = k_t*psi_sgm_ref - k*psi_sgm + self.u_i
        # Error (scaled, corresponds to the leakage flux linkage)
        self.e = psi_sgm_ref - psi_sgm

        return self.u_s_ref

    def update(self, u_s_ref_lim, w_s):
        """
        Update the integral state.

        Parameters
        ----------
        u_s_ref_lim : complex
            Limited voltage reference.
        w_s : float
            Angular stator frequency.

        """
        k_i = self.alpha_c*(self.alpha_c + 1j*w_s)
        k_t = self.alpha_c
        self.u_i += self.T_s*k_i*(self.e + (u_s_ref_lim - self.u_s_ref)/k_t)


# %%
class SensorlessObserver:
    """
    Sensorless reduced-order flux observer.

    This observer corresponds to [3]_. The observer gain decouples the
    electrical and mechanical dynamics and allows placing the poles of the
    corresponding linearized estimation error dynamics. This implementation
    operates in estimated rotor flux coordinates.

    Parameters
    ----------
    pars : InductionMotorVectorCtrlPars
        Control parameters.

    Notes
    -----
    This implementation corresponds to (26)-(30) in [3]_ with the choice
    c = w_s**2 in (17). The closed-loop poles, cf. (40), can still be
    affected via the coefficient b > 0.

    References
    ----------
    .. [3] Hinkkanen, Harnefors, Luomi, "Reduced-order flux observers with
       stator-resistance adaptation for speed-sensorless induction motor
       drives," IEEE Trans. Power Electron., 2010,
       https://doi.org/10.1109/TPEL.2009.2039650

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        self.T_s = pars.T_s
        self.R_s = pars.R_s
        self.R_R = pars.R_R
        self.L_sgm = pars.L_sgm
        self.L_M = pars.L_M
        self.alpha_o = pars.alpha_o
        self.zeta_inf = .2
        # Initial states
        self.theta_s, self.psi_R, self.i_s_old, self.w_m = 0, 0, 0, 0
        # Store for the update method to avoid recalculation
        self.dpsi_R, self.w_r = 0, 0

    def output(self, u_s, i_s, *_):
        """
        Compute the output.

        Parameters
        ----------
        u_s : complex
            Stator voltage in estimated rotor flux coordinates.
        i_s : complex
            Stator current in estimated rotor flux coordinates.

        Returns
        -------
        w_s : float
            Angular frequency of the rotor flux.

        """
        alpha = self.R_R/self.L_M

        # Observer gain (17) with c = w_s**2 (without the orthogonal projection
        # which is embedded into the state update)
        b = alpha + 2*self.zeta_inf*np.abs(self.w_m)
        g = b/(alpha - 1j*self.w_m)

        # Induced voltage from stator quantities, cf. (7)
        e_s = u_s - self.R_s*i_s - self.L_sgm*(i_s - self.i_s_old)/self.T_s
        # Induced voltage (8) from the rotor quantities
        e_r = self.R_R*i_s - (alpha - 1j*self.w_m)*self.psi_R

        # Angular frequency of the rotor flux vector
        den = self.psi_R + self.L_sgm*(i_s.real + g.imag*i_s.imag)
        if den > 0:
            w_s = (e_s.imag + g.imag*(e_r - e_s).real)/den
        else:
            w_s = self.w_m

        # Slip angular frequency (stored for the update method)
        self.w_r = e_r.imag/self.psi_R if self.psi_R > 0 else 0

        # Increment of the flux magnitude (stored for the update method)
        self.dpsi_R = ((1 - g.real)*(e_s.real + w_s*self.L_sgm*i_s.imag) +
                       g.real*e_r.real)

        return w_s

    def update(self, i_s, w_s):
        """Update the states for the next sampling period."""
        self.w_m += self.T_s*self.alpha_o*(w_s - self.w_r)
        self.psi_R += self.T_s*self.dpsi_R
        self.theta_s += self.T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.i_s_old = i_s


# %%
class Observer:
    """
    Sensored reduced-order flux observer.

    This reduced-order flux observer [4]_ uses the measured rotor speed. The
    selected default gain allows smooth transition from the current model at
    zero speed to the (damped) voltage model at higher speeds.

    Parameters
    ----------
    pars : InductionMotorVectorCtrlPars
        Control parameters.

    Notes
    -----
    This implementation places the pole in synchronous coordinates at::

        s = -R_R/L_M - g*abs(w_m) - 1j*(w_s - w_m)

    References
    ----------
    .. [4] Verghese, Sanders, “Observers for flux estimation in induction
       machines,” IEEE Trans. Ind. Electron., 1988,
       https://doi.org/10.1109/41.3067

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        self.T_s = pars.T_s
        self.R_s = pars.R_s
        self.R_R = pars.R_R
        self.L_sgm = pars.L_sgm
        self.alpha = pars.R_R/pars.L_M
        # Nonnegative gain for damping
        self.g = pars.g
        # Initial states
        self.theta_s, self.psi_R, self.i_s_old = 0, 0, 0
        # Store for the update method to avoid recalculation
        self.dpsi_R = 0

    def output(self, u_s, i_s, w_m):
        """
        Compute the output of the observer.

        Parameters
        ----------
        u_s : complex
            Stator voltage in estimated rotor flux coordinates.
        i_s : complex
            Stator current in estimated rotor flux coordinates.
        w_m : float
            Rotor angular speed (in electrical rad/s)

        Returns
        -------
        w_s : float
            Angular frequency of the rotor flux.

        """
        # The observer pole could be placed arbitrarily by means of the
        # observer gain k. The current model would be obtained using k = 1,
        # resulting in the pole at s = -R_R/L_M - 1j*(w_s - w_m). The pure
        # voltage model corresponds to k = 0, resulting in the marginally
        # stable pole at s = -1j*w_s.

        # This gain leads to s = -alpha - g*abs(w_m) - 1j*(w_s - w_m)
        k = 1 + self.g*np.abs(w_m)/(self.alpha - 1j*w_m)

        # Induced voltage from the stator quantities
        e_s = u_s - self.R_s*i_s - self.L_sgm*(i_s - self.i_s_old)/self.T_s
        # Induced voltage from the rotor quantities
        e_r = self.R_R*i_s - (self.alpha - 1j*w_m)*self.psi_R

        # Angular frequency of the rotor flux vector
        den = self.psi_R + self.L_sgm*((1 - k)*i_s).real
        v = (1 - k)*e_s + k*e_r
        w_s = v.imag/den if den > 0 else w_m

        # Increment of the flux magnitude (stored for the update method)
        self.dpsi_R = v.real + w_s*self.L_sgm*((1 - k)*i_s).imag

        return w_s

    def update(self, i_s, w_s):
        """Update the states for the next sampling period."""
        self.psi_R += self.T_s*self.dpsi_R
        self.theta_s += self.T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.i_s_old = i_s



"""Simulation environment."""

import numpy as np
from scipy.integrate import solve_ivp
from scipy.io import savemat
from motulator.helpers import abc2complex


# %%
class Delay:
    """
    Computational delay.

    This models the computational delay as a ring buffer.

    Parameters
    ----------
    length : int, optional
        Length of the buffer in samples. The default is 1.

    """

    def __init__(self, length=1, elem=3):
        self.data = length*[elem*[0]]  # Creates a zero list

    def __call__(self, u):
        """
        Delay the input.

        Parameters
        ----------
        u : array_like, shape (elem,)
            Input array.

        Returns
        -------
        array_like, shape (elem,)
            Output array.

        """
        # Add the latest value to the end of the list
        self.data.append(u)
        # Pop the first element and return it
        return self.data.pop(0)


# %%
class CarrierCmp:
    """
    Carrier comparison.

    This computes the the switching states and their durations based on the
    duty ratios. Instead of searching for zero crossings, the switching
    instants are explicitly computed in the beginning of each sampling period,
    allowing faster simulations.

    Parameters
    ----------
    N : int, optional
        Amount of the counter quantization levels. The default is 2**12.
    return_complex : bool, optional
        Complex switching state space vectors are returned if True. Otherwise
        phase switching states are returned. The default is True.

    Examples
    --------
    >>> from motulator.simulation import CarrierCmp
    >>> carrier_cmp = CarrierCmp(return_complex=False)
    >>> # First call gives rising edges
    >>> t_steps, q_abc = carrier_cmp(1e-3, [.4, .2, .8])
    >>> # Durations of the switching states
    >>> t_steps
    array([0.00019995, 0.00040015, 0.00019995, 0.00019995])
    >>> # Switching states
    >>> q_abc
    array([[0, 0, 0],
           [0, 0, 1],
           [1, 0, 1],
           [1, 1, 1]])
    >>> # Second call gives falling edges
    >>> t_steps, q_abc = carrier_cmp(.001, [.4, .2, .8])
    >>> t_steps
    array([0.00019995, 0.00019995, 0.00040015, 0.00019995])
    >>> q_abc
    array([[1, 1, 1],
           [1, 0, 1],
           [0, 0, 1],
           [0, 0, 0]])
    >>> # Sum of the step times equals T_s
    >>> np.sum(t_steps)
    0.001
    >>> # 50% duty ratios in all phases
    >>> t_steps, q_abc = carrier_cmp(1e-3, [.5, .5, .5])
    >>> t_steps
    array([0.0005, 0.    , 0.    , 0.0005])
    >>> q_abc
    array([[0, 0, 0],
           [0, 0, 0],
           [0, 0, 0],
           [1, 1, 1]])

    """

    def __init__(self, N=2**12, return_complex=True):
        self.N = N
        self.return_complex = return_complex
        self.rising_edge = True  # Stores the carrier direction

    def __call__(self, T_s, d_abc):
        """
        Compute the switching state durations and vectors.

        Parameters
        ----------
        T_s : float
            Half carrier period.
        d_abc : array_like of floats, shape (3,)
            Duty ratios in the range [0, 1].

        Returns
        -------
        t_steps : ndarray, shape (4,)
            Switching state durations, `[t0, t1, t2, t3]`.
        q : complex ndarray, shape (4,)
            Switching state vectors, `[q0, q1, q2, q3]`, where `q1` and `q2`
            are active vectors.

        Notes
        -----
        No switching (e.g. `d_a == 0` or `d_a == 1`) or simultaneous switchings
        (e.g. `d_a == d_b`) lead to zeroes in `t_steps`.

        """
        # Quantize the duty ratios to N levels
        d_abc = np.round(self.N*np.asarray(d_abc))/self.N

        # Assume falling edge and compute the normalized switching instants:
        t_n = np.append(0, np.sort(d_abc))
        # Compute the correponding switching states:
        q_abc = (t_n[:, np.newaxis] < d_abc).astype(int)

        # Durations of switching states
        t_steps = T_s*np.diff(t_n, append=1)

        # Flip the sequence if rising edge
        if self.rising_edge:
            t_steps = np.flip(t_steps)
            q_abc = np.flipud(q_abc)

        # Change the carrier direction for the next call
        self.rising_edge = not self.rising_edge

        return ((t_steps, abc2complex(q_abc.T)) if self.return_complex else
                (t_steps, q_abc))


# %%
def zoh(T_s, d_abc):
    """
    Zero-order hold of the duty ratios over the sampling period.

    Parameters
    ----------
    T_s : float
        Sampling period.
    d_abc : array_like of floats, shape (3,)
        Duty ratios in the range [0, 1].

    Returns
    -------
    t_steps : ndarray, shape (1,)
        Sampling period as an array compatible with the solver.
    q : complex ndarray, shape (1,)
        Duty ratio vector as an array compatible with the solver.

    """
    # Shape the output arrays to be compatible with the solver
    t_steps = np.array([T_s])
    q = np.array([abc2complex(d_abc)])
    return t_steps, q


# %%
class Simulation:
    """
    Simulation environment.

    Each simulation object has a system model object and a controller object.

    Parameters
    ----------
    mdl : InductionMotorDrive | SynchronousMotorDrive
        Continuous-time system model.
    ctrl : Ctrl
        Discrete-time controller.
    delay : int, optional
        Amount of computational delays. The default is 1.
    pwm : bool, optional
        Enable carrier comparison. The default is False.

    """

    def __init__(self, mdl=None, ctrl=None, delay=1, pwm=False):
        self.mdl = mdl
        self.ctrl = ctrl
        self.delay = Delay(delay)
        if pwm:
            self.pwm = CarrierCmp()
        else:
            self.pwm = zoh

    def simulate(self, t_stop=1, max_step=np.inf):
        """
        Solve the continuous-time model and call the discrete-time controller.

        Parameters
        ----------
        t_stop : float, optional
            Simulation stop time. The default is 1.
        max_step : float, optional
            Max step size of the solver. The default is inf.

        Notes
        -----
        Other options of solve_ivp could be easily changed if needed, but, for
        simplicity, only max_step is included as an option of this method.

        """
        try:
            self.simulation_loop(t_stop, max_step)
        except FloatingPointError:
            print(f'Invalid value encountered at {self.mdl.t0:.2f} seconds.')
        # Call the post-processing functions
        self.mdl.post_process()
        self.ctrl.post_process()

    @np.errstate(invalid='raise')
    def simulation_loop(self, t_stop, max_step):
        """Run the main simulation loop."""
        while self.mdl.t0 <= t_stop:

            # Run the digital controller
            T_s, d_abc_ref = self.ctrl(self.mdl)

            # Computational delay model
            d_abc = self.delay(d_abc_ref)

            # Carrier comparison
            t_steps, q = self.pwm(T_s, d_abc)

            # Loop over the sampling period T_s
            for i, t_step in enumerate(t_steps):

                if t_step > 0:
                    # Update the switching state
                    self.mdl.conv.q = q[i]

                    # Get initial values
                    x0 = self.mdl.get_initial_values()

                    # Integrate over t_span
                    t_span = (self.mdl.t0, self.mdl.t0 + t_step)
                    sol = solve_ivp(self.mdl.f, t_span, x0, max_step=max_step)

                    # Set the new initial values (last points of the solution)
                    t0_new, x0_new = t_span[-1], sol.y[:, -1]
                    self.mdl.set_initial_values(t0_new, x0_new)

                    # Save the solution
                    sol.q = len(sol.t)*[self.mdl.conv.q]
                    self.mdl.save(sol)

    def save_mat(self, name='sim'):
        """
        Save the simulation data into MATLAB .mat files.

        Parameters
        ----------
        name : str, optional
            Name for the simulation instance. The default is 'sim'.

        """
        savemat(name + '_mdl_data' + '.mat', self.mdl.data)
        savemat(name + '_ctrl_data' + '.mat', self.ctrl.data)



"""Continuous-time models for mechanical subsystems."""


# %%
class Mechanics:
    """
    Mechanics subsystem.

    This models an equation of motion for stiff mechanics.

    Parameters
    ----------
    J : float
        Total moment of inertia.
    tau_L_w : function
        Load torque as function of speed, `tau_L_w(w_M)`. For example,
        tau_L_w = b*w_M, where b is the viscous friction coefficient.
    tau_L_t : function
        Load torque as a function of time, `tau_L_t(t)`.

    """

    def __init__(
            self, J=.015, tau_L_w=lambda w_M: 0*w_M, tau_L_t=lambda t: 0*t):
        self.J = J
        self.tau_L_t = tau_L_t
        self.tau_L_w = tau_L_w
        # Initial values
        self.w_M0, self.theta_M0 = 0, 0

    def f(self, t, w_M, tau_M):
        """
        Compute the state derivatives.

        Parameters
        ----------
        t : float
            Time.
        w_M : float
            Rotor angular speed (in mechanical rad/s).
        tau_M : float
            Electromagnetic torque.

        Returns
        -------
        list, length 2
            Time derivatives of the state vector.

        """
        # Total load torque
        tau_L = self.tau_L_w(w_M) + self.tau_L_t(t)
        # Time derivatives
        dw_M = (tau_M - tau_L)/self.J
        dtheta_M = w_M
        return [dw_M, dtheta_M]

    def meas_speed(self):
        """
        Measure the rotor speed.

        This returns the rotor speed at the end of the sampling period.

        Returns
        -------
        w_M0 : float
            Rotor angular speed (in mechanical rad/s).

        """
        # The quantization noise of an incremental encoder could be modeled
        # by means of the rotor angle, to be done later.
        return self.w_M0

    def meas_position(self):
        """
        Measure the rotor angle.

        This returns the rotor angle at the end of the sampling period.

        Returns
        -------
        theta_M0 : float
            Rotor angle (in mechanical rad).

        """
        return self.theta_M0


# %%
class MechanicsTwoMass(Mechanics):
    """
    Two-mass mechanics subsystem.

    This models an equation of motion for two-mass mechanics.

    Parameters
    ----------
    J_M : float
        Moment of inertia of the motor.
    J_L : float
        Moment of inertia of the load.
    K_S : float
        Torsional stiffness of the shaft.
    C_S : float
        Torsional damping of the shaft.
    tau_L_w : function
        Load torque as function of the load speed, `tau_L_w(w_L)`. For example,
        tau_L_w = b*w_L, where b is the viscous friction coefficient.
    tau_L_t : function
        Load torque as a function of time, `tau_L_t(t)`.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
            self,
            J_M=.005,
            J_L=.005,
            K_S=700.,
            C_S=.13,
            tau_L_w=lambda w_M: 0*w_M,
            tau_L_t=lambda t: 0*t):
        # pylint: disable=too-many-arguments
        # pylint: disable=super-init-not-called
        self.J_M = J_M
        self.J_L = J_L
        self.K_S = K_S
        self.C_S = C_S
        self.tau_L_t = tau_L_t
        self.tau_L_w = tau_L_w
        # Initial values
        self.w_M0, self.theta_M0, self.w_L0, self.theta_ML0 = 0, 0, 0, 0

    def f(self, t, w_M, w_L, theta_ML, tau_M):
        # pylint: disable=too-many-arguments
        # pylint: disable=arguments-differ
        """
        Compute the state derivatives.

        Parameters
        ----------
        t : float
            Time.
        w_M : float
            Rotor angular speed (in mechanical rad/s).
        w_L : float
            Load angular speed (in mechanical rad/s).
        theta_ML : float
            Twist angle, theta_M - theta_L (in mechanical rad).
        tau_M : float
            Electromagnetic torque.

        Returns
        -------
        list, length 4
            Time derivatives of the state vector.

        """
        # Total load torque
        tau_L = self.tau_L_w(w_L) + self.tau_L_t(t)
        # Shaft torque
        tau_S = self.K_S*theta_ML + self.C_S*(w_M - w_L)
        # Time derivatives
        dw_M = (tau_M - tau_S)/self.J_M
        dtheta_M = w_M
        dw_L = (tau_S - tau_L)/self.J_L
        dtheta_ML = w_M - w_L

        return [dw_M, dtheta_M, dw_L, dtheta_ML]

    def meas_load_speed(self):
        """
        Measure the load speed.

        This returns the load speed at the end of the sampling period.

        Returns
        -------
        w_L0 : float
            Load angular speed (in mechanical rad/s).

        """
        return self.w_L0

    def meas_load_position(self):
        """
        Measure the load angle.

        This returns the load angle at the end of the sampling period.

        Returns
        -------
        theta_L0 : float
            Rotor angle (in mechanical rad).

        """
        theta_L0 = self.theta_M0 - self.theta_ML0
        return theta_L0



"""
Continuous-time models for induction motors.

Peak-valued complex space vectors are used. The space vector models are
implemented in stator coordinates. The default values correspond to a 2.2-kW
induction motor.

"""
import numpy as np
from motulator.helpers import complex2abc


# %%
class InductionMotor:
    """
    Γ-equivalent model of an induction motor.

    An induction motor is modeled using the Γ-equivalent model [1]_. The model
    is implemented in stator coordinates. The flux linkages are used as state
    variables.

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance.
    R_r : float
        Rotor resistance.
    L_ell : float
        Leakage inductance.
    L_s : float
        Stator inductance.

    Notes
    -----
    The Γ model is chosen here since it can be extended with the magnetic
    saturation model in a staightforward manner. If the magnetic saturation is
    omitted, the Γ model is mathematically identical to the inverse-Γ and T
    models [1]_.

    References
    ----------
    .. [1] Slemon, "Modelling of induction machines for electric drives," IEEE
       Trans. Ind. Appl., 1989, https://doi.org/10.1109/28.44251.

    """

    def __init__(self, n_p=2, R_s=3.7, R_r=2.5, L_ell=.023, L_s=.245):
        # pylint: disable=too-many-arguments
        self.n_p = n_p
        self.R_s, self.R_r = R_s, R_r
        self.L_ell, self.L_s = L_ell, L_s
        # Initial values
        self.psi_ss0, self.psi_rs0 = 0j, 0j

    def currents(self, psi_ss, psi_rs):
        """
        Compute the stator and rotor currents.

        Parameters
        ----------
        psi_ss : complex
            Stator flux linkage.
        psi_rs : complex
            Rotor flux linkage.

        Returns
        -------
        i_ss : complex
            Stator current.
        i_rs : complex
            Rotor current.

        """
        i_rs = (psi_rs - psi_ss)/self.L_ell
        i_ss = psi_ss/self.L_s - i_rs

        return i_ss, i_rs

    def magnetic(self, psi_ss, psi_rs):
        """
        Magnetic model.

        Parameters
        ----------
        psi_ss : complex
            Stator flux linkage.
        psi_rs : complex
            Rotor flux linkage.

        Returns
        -------
        i_ss : complex
            Stator current.
        i_rs : complex
            Rotor current.
        tau_M : float
            Electromagnetic torque.

        """
        i_ss, i_rs = self.currents(psi_ss, psi_rs)
        tau_M = 1.5*self.n_p*np.imag(i_ss*np.conj(psi_ss))

        return i_ss, i_rs, tau_M

    def f(self, psi_ss, psi_rs, u_ss, w_M):
        """
        Compute the state derivatives.

        Parameters
        ----------
        psi_ss : complex
            Stator flux linkage.
        psi_rs : complex
            Rotor flux linkage.
        u_ss : complex
            Stator voltage.
        w_M : float
            Rotor angular speed (in mechanical rad/s).

        Returns
        -------
        complex list, length 2
            Time derivative of the state vector, [dpsi_ss, dpsi_rs]
        i_ss : complex
            Stator current.
        tau_M : float
            Electromagnetic torque.

        Notes
        -----
        In addition to the state derivatives, this method also returns the
        output signals (stator current `i_ss` and torque `tau_M`) needed for
        interconnection with other subsystems. This avoids overlapping
        computation in simulation.

        """
        i_ss, i_rs, tau_M = self.magnetic(psi_ss, psi_rs)
        dpsi_ss = u_ss - self.R_s*i_ss
        dpsi_rs = -self.R_r*i_rs + 1j*self.n_p*w_M*psi_rs

        return [dpsi_ss, dpsi_rs], i_ss, tau_M

    def meas_currents(self):
        """
        Measure the phase currents at the end of the sampling period.

        Returns
        -------
        i_s_abc : 3-tuple of floats
            Phase currents.

        """
        # Stator current space vector in stator coordinates
        i_ss, _ = self.currents(self.psi_ss0, self.psi_rs0)
        # Phase currents
        i_s_abc = complex2abc(i_ss)  # + noise + offset ...
        return i_s_abc


# %%
class InductionMotorSaturated(InductionMotor):
    """
    Γ-equivalent model of an induction motor model with main-flux saturation.

    This extends the InductionMotor class with a main-flux magnetic saturation
    model [2]_::

        L_s(psi_ss) = L_su/(1 + (beta*abs(psi_ss)**S)

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance.
    R_r : float
        Rotor resistance.
    L_ell : float
        Leakage inductance.
    L_su : float
        Unsaturated stator inductance.
    beta : float
        Positive coefficient.
    S : float
        Positive coefficient.

    References
    ----------
    .. [2] Qu, Ranta, Hinkkanen, Luomi, "Loss-minimizing flux level control of
       induction motor drives," IEEE Trans. Ind. Appl., 2012,
       https://doi.org/10.1109/TIA.2012.2190818

    """

    def __init__(
            self,
            n_p=2,
            R_s=3.7,
            R_r=2.5,
            L_ell=.023,
            L_su=.34,
            beta=.84,
            S=7):
        # pylint: disable=too-many-arguments
        super().__init__(n_p=n_p, R_s=R_s, R_r=R_r, L_ell=L_ell)
        # Saturation model
        self.L_s = lambda psi: L_su/(1. + (beta*np.abs(psi))**S)

    def currents(self, psi_ss, psi_rs):
        """Override the base class method."""
        # Saturated value of the stator inductance.
        L_s = self.L_s(psi_ss)
        # Currents
        i_rs = (psi_rs - psi_ss)/self.L_ell
        i_ss = psi_ss/L_s - i_rs
        return i_ss, i_rs


# %%
class InductionMotorInvGamma(InductionMotor):
    """
    Inverse-Γ model of an induction motor.

    This extends the InductionMotor class (based on the Γ model) by providing
    an interface for the inverse-Γ model parameters. Linear magnetics are
    assumed. If magnetic saturation is to be modeled, the Γ model is preferred.

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance.
    R_R : float
        Rotor resistance.
    L_sgm : float
        Leakage inductance.
    L_M : float
        Magnetizing inductance.

    """

    def __init__(self, n_p=2, R_s=3.7, R_R=2.1, L_sgm=.021, L_M=.224):
        # pylint: disable=too-many-arguments, disable=super-init-not-called
        # Convert the inverse-Γ parameters to the Γ parameters
        gamma = L_M/(L_M + L_sgm)  # Magnetic coupling factor
        self.n_p = n_p
        self.R_s = R_s
        self.L_s = L_M + L_sgm
        self.L_ell = L_sgm/gamma
        self.R_r = R_R/gamma**2
        # Initial values
        self.psi_ss0 = 0j
        self.psi_rs0 = 0j  # self.psi_rs0 = self.psi_Rs0/gamma



"""
Continuous-time models for synchronous motors.

The motor models can be parametrized to represent permanent-magnet synchronous
motors and synchronous reluctance motors. Peak-valued complex space vectors are
used.

"""
import numpy as np
from scipy.optimize import minimize_scalar
from scipy.interpolate import LinearNDInterpolator
from motulator.helpers import complex2abc


# %%
class SynchronousMotor:
    """
    Synchronous motor model.

    This models a synchronous motor in rotor coordinates. The stator flux
    linkage is the state variable. The default values correspond to a 2.2-kW
    permanent-magnet synchronous motor.

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance.
    L_d : float
        d-axis inductance.
    L_q : float
        q-axis inductance.
    psi_f : float
        PM-flux linkage.
    mech : Mechanics
        Model of the mechanical subsystem, needed only for the coordinate
        transformation in the measure_currents method.

    """

    def __init__(
            self, n_p=3, R_s=3.6, L_d=.036, L_q=.051, psi_f=.545, mech=None):
        # pylint: disable=too-many-arguments
        self.n_p, self.R_s = n_p, R_s
        self.L_d, self.L_q, self.psi_f = L_d, L_q, psi_f
        # Initial value
        self.psi_s0 = self.psi_f + 0j
        # For the coordinate transformation
        self._mech = mech

    def current(self, psi_s):
        """
        Compute the stator current.

        Parameters
        ----------
        psi_s : complex
            Stator flux linkage.

        Returns
        -------
        i_s : complex
            Stator current.

        """
        i_s = (psi_s.real - self.psi_f)/self.L_d + 1j*psi_s.imag/self.L_q
        return i_s

    def magnetic(self, psi_s):
        """
        Magnetic model.

        Parameters
        ----------
        psi_s : complex
            Stator flux linkage.

        Returns
        -------
        i_s : complex
            Stator current.
        tau_M : float
            Electromagnetic torque.

        """
        i_s = self.current(psi_s)
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_s))
        return i_s, tau_M

    def f(self, psi_s, u_s, w_M):
        """
        Compute the state derivative.

        Parameters
        ----------
        psi_s : complex
            Stator flux linkage.
        u_s : complex
            Stator voltage.
        w_M : float
            Rotor angular speed (in mechanical rad/s).

        Returns
        -------
        dpsi_s : complex list
            Time derivative of the stator flux linkage.
        i_s : complex
            Stator current.
        tau_M : float
            Electromagnetic torque.

        Notes
        -----
        In addition to the state derivative, this method also returns the
        output signals (stator current `i_ss` and torque `tau_M`) needed for
        interconnection with other subsystems. This avoids overlapping
        computation in simulation.

        """
        i_s, tau_M = self.magnetic(psi_s)
        dpsi_s = u_s - self.R_s*i_s - 1j*self.n_p*w_M*psi_s
        return [dpsi_s], i_s, tau_M

    def meas_currents(self):
        """
        Measure the phase currents at the end of the sampling period.

        Returns
        -------
        i_s_abc : 3-tuple of floats
            Phase currents.

        """
        i_s0 = self.current(self.psi_s0)
        theta_m0 = self.n_p*self._mech.theta_M0
        i_s_abc = complex2abc(np.exp(1j*theta_m0)*i_s0)
        return i_s_abc


# %%
class SynchronousMotorSaturated(SynchronousMotor):
    """
    Model of a saturated synchronous motor.

    This extends the SynchronousMotor class with an analytical saturation
    model [1]_, [2]_. The permanent magnets (PMs) are assumed to be along the
    d-axis. The default values correspond to a 6.7-kW synchronous reluctance
    motor.

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance.
    i_f : float
        Constant current corresponding to the magnetomotive force (MMF) of PMs.
        In the magnetically linear case, `i_f = psi_f/L_d`.
    a_d0 : float
        Nonnegative parameter of the saturation model. In the magnetically
        linear case, `a_d0 = 1/L_d`.
    a_q0 : float
        Nonnegative parameter of the saturation model. In the magnetically
        linear case, `a_q0 = 1/L_q`.
    a_dd : float
        Nonnegative constant defining the d-axis self-saturation together with
        `S`. In the magnetically linear case, `a_dd = 0`.
    a_qq : float
        Nonnegative constant defining the q-axis self-saturation together with
        `T`. In the magnetically linear case, `a_qq = 0`.
    a_dq : float
        Nonnegative constant defining the cross-saturation together with `U`
        and `V`. In the magnetically linear case, `a_dq = 0`.
    S : float
        Nonnegative constant defining the d-axis self-saturation.
    T : float
        Nonnegative constant defining the q-axis self-saturation.
    U : float
        Nonnegative constant defining the cross-saturation.
    V : float
        Nonnegative constant defining the cross-saturation.
    mech : Mechanics
        Model of the mechanical subsystem, needed only for the coordinate
        transformation in the measure_currents method.

    Notes
    -----
    The magnetomotive force (MMF) of the PMs is modeled using constant current
    source `i_f` on the d-axis [3]_. Correspondingly, this approach assumes
    that the MMFs of the d-axis current and of the PMs are in series. This
    model cannot capture the desaturation phenomenon of thin iron ribs [4]_.
    For such motors, look-up tables can be used.

    References
    ----------
    .. [1] Hinkkanen, Pescetto, Mölsä, Saarakkala, Pellegrino, Bojoi,
       “Sensorless self-commissioning of synchronous reluctance motors at
       standstill without rotor locking, ”IEEE Trans. Ind. Appl., 2017,
       https://doi.org/10.1109/TIA.2016.2644624

    .. [2] Awan, Song, Saarakkala, Hinkkanen, "Optimal torque control of
       saturated synchronous motors: plug-and-play method," IEEE Trans. Ind.
       Appl., 2018, https://doi.org/10.1109/TIA.2018.2862410

    .. [3] Jahns, Kliman, Neumann, “Interior permanent-magnet synchronous
       motors for adjustable-speed drives,” IEEE Trans. Ind. Appl., 1986,
       https://doi.org/10.1109/TIA.1986.4504786

    .. [4] Armando, Guglielmi, Pellegrino, Pastorelli, Vagati, "Accurate
       modeling and performance analysis of IPM-PMASR motors," IEEE Trans. Ind.
       Appl., 2009, https://doi.org/10.1109/TIA.2008.2009493

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        n_p=2,
        R_s=.54,
        i_f=0,
        a_d0=17.4,
        a_q0=52.1,
        a_dd=373.,
        a_qq=658.,
        a_dq=1120.,
        S=5,
        T=1,
        U=1,
        V=0,
        mech=None,
    ):
        # pylint: disable=too-many-arguments, disable=super-init-not-called
        self.n_p, self.R_s = n_p, R_s
        self.i_f, self.a_d0, self.a_q0 = i_f, a_d0, a_q0
        self.a_dd, self.a_qq, self.a_dq = a_dd, a_qq, a_dq
        self.S, self.T, self.U, self.V = S, T, U, V

        # Initial value of the stator flux
        if i_f == 0:
            # No magnets
            self.psi_s0 = 0j
        else:
            # Solve the stator flux caused by the magnets @ i_s = 0
            res = minimize_scalar(
                lambda psi_d: np.abs(
                    (a_d0 + a_dd*np.abs(psi_d)**S)*psi_d - i_f))
            self.psi_s0 = complex(res.x)
            print(self.psi_s0)

        # For the coordinate transformation
        self._mech = mech

    def current(self, psi_s):
        """Override the base class method."""
        abs_psi_d = np.abs(psi_s.real)
        abs_psi_q = np.abs(psi_s.imag)

        G_dd = self.a_d0 + self.a_dd*abs_psi_d**self.S
        G_dq = self.a_dq/(self.V + 2)*abs_psi_d**self.U*abs_psi_q**(self.V + 2)
        G_qq = self.a_q0 + self.a_qq*abs_psi_q**self.T
        G_qd = self.a_dq/(self.U + 2)*abs_psi_d**(self.U + 2)*abs_psi_q**self.V

        i_d = (G_dd + G_dq)*psi_s.real - self.i_f
        i_q = (G_qq + G_qd)*psi_s.imag
        i_s = i_d + 1j*i_q

        return i_s


# %%
class SynchronousMotorSaturatedLUT(SynchronousMotor):
    """
    Look-up-table-based model of a saturated synchronous motor.

    This extends the SynchronousMotor class with a saturation model, where the
    stator current depends on the stator flux linkage. The coordinates assume
    the PMSM convention, i.e., that the PM flux is along the d-axis.
    Unstructured flux map data can be used.

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance.
    psi_s_data : ndarray of complex
        Stator flux data points for creating the interpolant.
    i_s_data : ndarray of complex
        Stator current data values for creating the interpolant.
    mech : Mechanics
        Model of the mechanical subsystem, needed only for the coordinate
        transformation in the measure_currents method.

    """

    # pylint: disable=too-many-arguments, disable=super-init-not-called
    def __init__(
            self, n_p=2, R_s=.20, psi_s_data=None, i_s_data=None, mech=None):

        self.n_p, self.R_s = n_p, R_s

        # Create the interpolant
        self.i_s = LinearNDInterpolator(
            list(zip(psi_s_data.real, psi_s_data.imag)), i_s_data)

        # Solve the PM flux for the initial value of the stator flux
        res = minimize_scalar(
            lambda psi_d: np.abs(self.i_s(psi_d, 0)),
            bounds=(0, np.max(psi_s_data.real)),
            method='bounded')
        self.psi_s0 = complex(res.x)

        # For the coordinate transformation
        self._mech = mech

    def current(self, psi_s):
        """Override the base class method."""
        # Return the current as function of the flux linkage
        return self.i_s(psi_s.real, psi_s.imag)



"""
Torque characteristics for synchronous machines.

This contains computation and plotting of torque characteristics for
synchronous machines, including the MTPA and MTPV loci [1]_. The methods can be
used to define look-up tables for control and to analyze the characteristics.
In this version, the magnetic saturation is omitted.

References
----------
.. [1] Morimoto, Takeda, Hirasa, Taniguchi, "Expansion of operating limits for
   permanent magnet motor by current vector control considering inverter
   capacity," IEEE Trans. Ind. Appl., 1990,
   https://doi.org/10.1109/28.60058

"""
from sys import float_info
import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from cycler import cycler
from motulator.helpers import Bunch

plt.rcParams['axes.prop_cycle'] = cycler(color='brgcmyk')
plt.rcParams['lines.linewidth'] = 1.
plt.rcParams.update({"text.usetex": False})  # Disable LaTeX in plots


# %%
class TorqueCharacteristics:
    """
    Compute MTPA and MTPV loci based on the motor parameters.

    The magnetic saturation is omitted.

    Parameters
    ----------
    pars : data object
        Motor parameters.

    """

    def __init__(self, pars):
        self.n_p = pars.n_p
        self.L_d = pars.L_d
        self.L_q = pars.L_q
        self.psi_f = pars.psi_f
        try:
            self.psi_s_min = pars.psi_s_min
        except AttributeError:
            self.psi_s_min = None
        try:
            self.psi_s_max = pars.psi_s_max
        except AttributeError:
            self.psi_s_max = None

    def torque(self, psi_s):
        """
        Compute the torque as a function of the stator flux linkage.

        Parameters
        ----------
        psi_s : complex
            Stator flux.

        Returns
        -------
        tau_M : float
            Electromagnetic torque.

        """
        i_s = self.current(psi_s)
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_s))

        return tau_M

    def current(self, psi_s):
        """
        Compute the stator current as a function of the stator flux linkage.

        Parameters
        ----------
        psi_s : complex
            Stator flux linkage.

        Returns
        -------
        i_s : complex
            Stator current.

        """
        i_s = (psi_s.real - self.psi_f)/self.L_d + 1j*psi_s.imag/self.L_q

        return i_s

    def flux(self, i_s):
        """
        Compute the stator flux linkage as a function of the current.

        Parameters
        ----------
        i_s : complex
            Stator current.

        Returns
        -------
        psi_s : complex
            Stator flux linkage.

        """
        psi_s = self.L_d*i_s.real + self.psi_f + 1j*self.L_q*i_s.imag

        return psi_s

    def mtpa(self, abs_i_s):
        """
        Compute the MTPA stator current angle.

        Parameters
        ----------
        abs_i_s : float
            Stator current magnitude.

        Returns
        -------
        beta : float
            MTPA angle of the stator current vector.

        """
        # Replace zeros with epsilon
        abs_i_s = (abs_i_s > 0)*abs_i_s + (abs_i_s <= 0)*float_info.epsilon

        if self.psi_f == 0:
            # SyRM (d-axis aligned with the maximum inductance)
            beta = .25*np.pi
        elif self.L_d == self.L_q:
            # Nonsalient machine
            beta = .5*np.pi
        else:
            # Salient machine
            a = self.psi_f/((self.L_q - self.L_d)*abs_i_s)

            if self.L_q > self.L_d:
                beta = np.arccos(.25*(a - np.sqrt(a**2 + 8)))
            else:
                beta = np.arccos(.25*(a + np.sqrt(a**2 + 8)))

        return beta

    def mtpv(self, abs_psi_s):
        """
        Compute the MTPV stator flux angle.

        Parameters
        ----------
        abs_psi_s : float
            Stator flux magnitude.

        Returns
        -------
        delta : float
            MTPV angle of the stator flux vector.

        """
        # Replace zeros with epsilon
        abs_psi_s = ((abs_psi_s > 0)*abs_psi_s +
                     (abs_psi_s <= 0)*float_info.epsilon)

        if self.psi_f == 0:
            # SyRM (d-axis aligned with the maximum inductance)
            delta = .25*np.pi
        elif self.L_d == self.L_q:
            # Nonsalient machine
            delta = .5*np.pi
        else:
            # Salient machine
            a = self.L_q/(self.L_q - self.L_d)*self.psi_f/abs_psi_s

            if self.L_q > self.L_d:
                delta = np.arccos(.25*(a - np.sqrt(a**2 + 8)))
            else:
                delta = np.arccos(.25*(a + np.sqrt(a**2 + 8)))

        return delta

    def mtpv_current(self, abs_i_s):
        """
        Compute the MTPV based on the current magnitude.

        This computes the MTPV based on the current magnitude, i.e., the
        intersection of the MTPV current locus and the current limit circle.
        This method is not necessary for computing the control look-up tables.
        It is used here to "cut" the MTPV characteristics at the desired
        current. Alternatively just a large enough maximum flux magnitude could
        be used.

        Parameters
        ----------
        abs_i_s : float
            Stator current magnitude.

        Returns
        -------
        i_s : complex
            MTPV stator current.

        """
        if self.psi_f == 0:
            # SyRM
            i_s = abs_i_s*np.exp(1j*(np.arctan(self.L_d/self.L_q)))
        elif self.psi_f/self.L_d < abs_i_s:
            if self.L_d == self.L_q:
                # Nonsalient machine
                i_sd = -self.psi_f/self.L_d
                i_sq = np.sqrt(abs_i_s**2 - i_sd**2)
                i_s = i_sd + 1j*i_sq
            else:
                # Salient machine
                k = self.L_q/(self.L_d - self.L_q)
                a = self.L_d**2 + self.L_q**2
                b = (2 + k)*self.psi_f*self.L_d
                c = (1 + k)*self.psi_f**2 - (self.L_q*abs_i_s)**2
                if self.L_q > self.L_d:
                    i_sd = .5*(-b - np.sqrt(b**2 - 4*a*c))/a
                else:
                    i_sd = .5*(-b + np.sqrt(b**2 - 4*a*c))/a
                i_sq = np.sqrt(abs_i_s**2 - i_sd**2)
                i_s = i_sd + 1j*i_sq
        else:
            # No MTPV for the given current magnitude
            i_s = np.nan

        return i_s

    def mtpa_locus(self, i_s_max=1, N=20):
        """
        Compute the MTPA locus.

        Parameters
        ----------
        i_s_max : float, optional
            Maximum stator current magnitude at which the locus is computed.
            The default is 1.
        N : int, optional
            Amount of points. The default is 20.

        Returns
        -------
        Bunch object with the following fields defined:
        psi_s : complex
            Stator flux.
        i_s : complex
            Stator current.
        tau_M : float
            Electromagnetic torque.
        abs_psi_s_vs_tau_M : interp1d object
            Stator flux magnitude as a function of the torque.
        i_sd_vs_tau_M : interp1d object
            d-axis current as a function of the torque.

        """
        # Current  magnitudes
        abs_i_s = np.linspace(0, i_s_max, N)

        # MTPA locus expressed with different quantities
        beta = self.mtpa(abs_i_s)
        i_s = abs_i_s*np.exp(1j*beta)

        if (self.psi_s_min is not None) and (self.psi_f == 0):
            # Minimum d-axis current for sensorless SyRM drives
            i_sd_min = self.psi_s_min/self.L_d
            i_s.real = ((i_s.real < i_sd_min)*i_sd_min +
                        (i_s.real >= i_sd_min)*i_s.real)

        psi_s = self.flux(i_s)
        tau_M = self.torque(psi_s)

        # Create an interpolant that can be used as a look-up table. If needed,
        # more interpolants can be easily added.
        abs_psi_s_vs_tau_M = interp1d(
            tau_M, np.abs(psi_s), fill_value='extrapolate')
        i_sd_vs_tau_M = interp1d(tau_M, i_s.real, fill_value='extrapolate')

        # Return the result as a bunch object
        return Bunch(
            psi_s=psi_s,
            i_s=i_s,
            tau_M=tau_M,
            abs_psi_s_vs_tau_M=abs_psi_s_vs_tau_M,
            i_sd_vs_tau_M=i_sd_vs_tau_M)

    def mtpv_locus(self, psi_s_max=1, i_s_max=None, N=20):
        """
        Compute the MTPV locus.

        Parameters
        ----------
        psi_s_max : float, optional
            Maximum stator flux magnitude at which the locus is computed. The
            default is 1.
        i_s_max : float, optional
            Maximum stator current magnitude at which the locus is computed.
            The default is None.
        N : int, optional
            Amount of points. The default is 20.

        Returns
        -------
        Bunch object with the following fields defined:
        psi_s : complex
            Stator flux.
        i_s : complex
            Stator current.
        tau_M : float
            Electromagnetic torque.
        tau_M_vs_abs_psi_s : interp1d object
            Torque as a function of the flux magnitude.

        """
        # If i_s_max is given, compute the corresponding MTPV stator flux
        if i_s_max:
            i_s_mtpv = self.mtpv_current(i_s_max)
            psi_s_max = np.abs(self.flux(i_s_mtpv))

        # Flux magnitudes
        abs_psi_s = np.linspace(0, psi_s_max, N)

        # MTPV locus expressed with different quantities
        delta = self.mtpv(abs_psi_s)
        psi_s = abs_psi_s*np.exp(1j*delta)
        i_s = self.current(psi_s)
        tau_M = self.torque(psi_s)

        # Create an interpolant that can be used as a look-up table. If needed,
        # more interpolants can be easily added.
        tau_M_vs_abs_psi_s = interp1d(np.abs(psi_s), tau_M, bounds_error=False)

        # Return the result as a bunch object
        return Bunch(
            psi_s=psi_s,
            i_s=i_s,
            tau_M=tau_M,
            tau_M_vs_abs_psi_s=tau_M_vs_abs_psi_s)

    def current_limit(self, i_s_max=1, gamma1=np.pi, gamma2=0, N=20):
        """
        Compute the current limit.

        Parameters
        ----------
        i_s_max : float, optional
            Current limit. The default is 1.
        gamma1 : float, optional
            Starting angle in radians. The default is 0.
        gamma2 : float, optional
            End angle in radians. The defauls in np.pi.
        N : int, optional
            Amount of points. The default is 20.

        Returns
        -------
        Bunch object with the following fields defined:
        psi_s : complex
            Stator flux.
        i_s : complex
            Stator current.
        tau_M : float
            Electromagnetic torque.
        tau_M_vs_abs_psi_s : interp1d object
            Torque as a function of the flux magnitude.

        """
        if np.isnan(gamma1):
            # No MTPV
            gamma1 = np.pi

        gamma = np.linspace(gamma1, gamma2, N)

        # MTPA locus expressed with different quantities
        i_s = i_s_max*np.exp(1j*gamma)
        psi_s = self.flux(i_s)
        tau_M = self.torque(psi_s)

        # Create an interpolant that can be used as a look-up table. If needed,
        # more interpolants can be easily added.
        tau_M_vs_abs_psi_s = interp1d(
            np.abs(psi_s),
            tau_M,
            bounds_error=False,
            fill_value=(tau_M[0], tau_M[-1]))

        # Return the result as a bunch object
        return Bunch(
            psi_s=psi_s,
            i_s=i_s,
            tau_M=tau_M,
            tau_M_vs_abs_psi_s=tau_M_vs_abs_psi_s)

    def mtpv_and_current_limits(self, i_s_max=1, N=20):
        """
        Merge the MTPV and current limits into a single interpolant.

        Parameters
        ----------
        i_s_max : float, optional
            Current limit. The default is 1.
        N : int, optional
            Amount of points. The default is 20.

        Returns
        -------
        Bunch object with the following fields defined:
        tau_M_vs_abs_psi_s : interp1d object
            Torque as a function of the flux magnitude.
        i_sd_vs_tau_M : interp1d object
            d-axis current as a function of the torque.

        """
        mtpa = self.mtpa_locus(i_s_max=i_s_max, N=N)
        mtpv = self.mtpv_locus(i_s_max=i_s_max, N=N)
        clim = self.current_limit(
            i_s_max=i_s_max,
            N=N,
            gamma1=np.angle(mtpv.i_s[-1]),
            gamma2=np.angle(mtpa.i_s[-1]))

        if np.isnan(mtpv.i_s).any():
            # No MTPV, only the current limit
            psi_s = clim.psi_s
            tau_M = clim.tau_M
            i_sd = clim.i_s.real
        else:
            # Concatenate the MTPV and current limits
            psi_s = np.concatenate((mtpv.psi_s, clim.psi_s))
            tau_M = np.concatenate((mtpv.tau_M, clim.tau_M))
            i_sd = np.concatenate((mtpv.i_s.real, clim.i_s.real))

        # Create an interpolant that can be used as a look-up table
        tau_M_vs_abs_psi_s = interp1d(
            np.abs(psi_s),
            tau_M,
            bounds_error=False,
            fill_value=(tau_M[0], tau_M[-1]))
        i_sd_vs_tau_M = interp1d(tau_M, i_sd, fill_value='extrapolate')

        # Return the result as a bunch object
        return Bunch(
            tau_M_vs_abs_psi_s=tau_M_vs_abs_psi_s, i_sd_vs_tau_M=i_sd_vs_tau_M)

    def plot_flux_loci(self, i_s_max, base, N=20):
        """
        Plot the stator flux linkage loci.

        Per-unit quantities are used.

        Parameters
        ----------
        i_s_max : float
            Maximum current at which the loci are evaluated.
        base : object
            Base values.
        N : int, optional
            Amount of points to be evaluated. The default is 20.

        """
        # Compute the characteristics
        mtpa = self.mtpa_locus(i_s_max=i_s_max, N=N)
        mtpv = self.mtpv_locus(i_s_max=i_s_max, N=N)
        clim = self.current_limit(
            i_s_max=i_s_max,
            N=N,
            gamma1=np.angle(mtpv.i_s[-1]),
            gamma2=np.angle(mtpa.i_s[-1]))

        # Plot the i_sd--i_sq current plane
        _, ax = plt.subplots()
        ax.plot(
            mtpa.psi_s.real/base.psi, mtpa.psi_s.imag/base.psi, label='MTPA')
        try:
            ax.plot(
                mtpv.psi_s.real/base.psi,
                mtpv.psi_s.imag/base.psi,
                label='MTPV')
        except AttributeError:
            pass
        ax.plot(
            clim.psi_s.real/base.psi,
            clim.psi_s.imag/base.psi,
            label='Const current')

        ax.legend()
        ax.set_xlabel(r'$\psi_\mathrm{sd}$ (p.u.)')
        ax.set_ylabel(r'$\psi_\mathrm{sq}$ (p.u.)')
        ax.set_ylim(0, None)
        ax.set_aspect('equal')

    def plot_current_loci(self, i_s_max, base, N=20):
        """
        Plot the current loci.

        Per-unit quantities are used.

        Parameters
        ----------
        i_s_max : float
            Maximum current at which the loci are evaluated.
        base : object
            Base values.
        N : int, optional
            Amount of points to be evaluated. The default is 20.

        """
        # Compute the characteristics
        mtpa = self.mtpa_locus(i_s_max=i_s_max, N=N)
        mtpv = self.mtpv_locus(i_s_max=i_s_max, N=N)
        clim = self.current_limit(
            i_s_max=i_s_max,
            N=N,
            gamma1=np.angle(mtpv.i_s[-1]),
            gamma2=np.angle(mtpa.i_s[-1]))

        # Plot the i_sd--i_sq current plane
        _, ax = plt.subplots()
        ax.plot(mtpa.i_s.real/base.i, mtpa.i_s.imag/base.i, label='MTPA')
        try:
            ax.plot(mtpv.i_s.real/base.i, mtpv.i_s.imag/base.i, label='MTPV')
        except AttributeError:
            pass
        ax.plot(
            clim.i_s.real/base.i, clim.i_s.imag/base.i, label='Const current')

        ax.set_xlabel(r'$i_\mathrm{sd}$ (p.u.)')
        ax.set_ylabel(r'$i_\mathrm{sq}$ (p.u.)')
        ax.legend()
        if self.psi_f == 0:
            ax.axis([0, i_s_max/base.i, 0, i_s_max/base.i])
        elif self.L_q > self.L_d:
            ax.axis([-i_s_max/base.i, 0, 0, i_s_max/base.i])
        else:
            ax.axis([-i_s_max/base.i, i_s_max/base.i, 0, i_s_max/base.i])
        ax.set_aspect('equal')

    def plot_torque_current(self, i_s_max, base, N=20):
        """
        Plot torque vs. current characteristics.

        Per-unit quantities are used.

        Parameters
        ----------
        i_s_max : float
            Maximum current at which the loci are evaluated.
        base : object
            Base values.
        N : int, optional
            Amount of points to be evaluated. The default is 20.

        """
        # Compute the characteristics
        mtpa = self.mtpa_locus(i_s_max=i_s_max, N=N)
        mtpv = self.mtpv_locus(i_s_max=i_s_max, N=N)
        clim = self.current_limit(
            i_s_max=i_s_max,
            N=N,
            gamma1=np.angle(mtpv.i_s[-1]),
            gamma2=np.angle(mtpa.i_s[-1]))

        # Plot i_sd vs. tau_M
        _, (ax1, ax2) = plt.subplots(2, 1)
        ax1.plot(mtpa.tau_M/base.tau, mtpa.i_s.real/base.i)
        try:
            ax1.plot(mtpv.tau_M/base.tau, mtpv.i_s.real/base.i)
        except AttributeError:
            pass
        ax1.plot(clim.tau_M/base.tau, clim.i_s.real/base.i)

        ax1.set_xlim(0, i_s_max/base.i)
        if self.psi_f == 0:
            ax1.set_ylim(0, None)
        elif self.L_q > self.L_d:
            ax1.set_ylim(None, 0)
        ax1.legend(['MTPA', 'MTPV', 'Const current'])
        ax1.set_ylabel(r'$i_\mathrm{sd}$ (p.u.)')

        # Plot i_sq vs. tau_M
        ax2.plot(mtpa.tau_M/base.tau, mtpa.i_s.imag/base.i)
        try:
            ax2.plot(mtpv.tau_M/base.tau, mtpv.i_s.imag/base.i)
        except TypeError:
            pass
        ax2.plot(clim.tau_M/base.tau, clim.i_s.imag/base.i)

        ax2.set_xlim(0, i_s_max/base.i)
        ax2.set_ylim(0, None)
        ax2.legend(['MTPA', 'MTPV', 'Const current'])
        ax2.set_ylabel(r'$i_\mathrm{sq}$ (p.u.)')
        ax2.set_xlabel(r'$\tau_\mathrm{M}$ (p.u.)')

    def plot_torque_flux(self, i_s_max, base, N=20):
        """
        Plot torque vs. flux magnitude characteristics.

        Per-unit quantities are used.

        Parameters
        ----------
        i_s_max : float
            Maximum current at which the loci are evaluated.
        base : object
            Base values.
        N : int, optional
            Amount of points to be evaluated. The default is 20.

        """
        # Compute the characteristics
        mtpa = self.mtpa_locus(i_s_max=i_s_max, N=N)
        mtpv = self.mtpv_locus(i_s_max=i_s_max, N=N)
        clim = self.current_limit(
            i_s_max=i_s_max,
            N=N,
            gamma1=np.angle(mtpv.i_s[-1]),
            gamma2=np.angle(mtpa.i_s[-1]))

        # Plot
        _, ax = plt.subplots(1, 1)
        ax.plot(np.abs(mtpa.psi_s)/base.psi, mtpa.tau_M/base.tau)
        try:
            ax.plot(np.abs(mtpv.psi_s)/base.psi, mtpv.tau_M/base.tau)
        except AttributeError:
            pass
        ax.plot(np.abs(clim.psi_s)/base.psi, clim.tau_M/base.tau)

        ax.legend(['MTPA', 'MTPV', 'Const current'])
        ax.set_xlabel(r'$\psi_\mathrm{s}$ (p.u.)')
        ax.set_ylabel(r'$\tau_\mathrm{m}$ (p.u.)')



"""
Observer-based V/Hz control for induction motor drives.

This implements the observer-based V/Hz control method described in [1]_. The
state-feedback control law is in the alternative form which uses an
intermediate stator current reference.

References
----------
.. [1] Tiitinen, Hinkkanen, Harnefors, "Stable and passive observer-based V/Hz
    control for induction motors," Proc. IEEE ECCE, Detroit, MI, Oct. 2022.

"""

# %%
from typing import Callable
from dataclasses import dataclass, field
import numpy as np
from motulator.control.common import Ctrl, PWM, RateLimiter
from motulator.helpers import abc2complex, Bunch


# %%
@dataclass
class InductionMotorObsVHzCtrlPars:
    """Control parameters."""

    # pylint: disable=too-many-instance-attributes
    # Speed reference (in electrical rad/s)
    w_m_ref: Callable[[float], float] = field(
        repr=False, default=lambda t: (t > .2)*(2*np.pi*50))

    # Control
    T_s: float = 250e-6
    psi_s_nom: float = 1.04  # 1 p.u.
    rate_limit: float = 2*np.pi*120
    i_s_max: float = 1.5*np.sqrt(2)*5
    alpha_f: float = 2*np.pi*1
    alpha_psi: float = 2*np.pi*20
    k_tau: float = 3.
    six_step: bool = False

    # Slip compensation
    slip_compensation: bool = True
    alpha_r: float = 2*np.pi*1

    # Observer
    alpha_o: float = 2*np.pi*40
    zeta_inf: float = .7

    # Motor parameter estimates (inverse-Γ model)
    R_s: float = 3.7
    R_R: float = 2.1
    L_sgm: float = .021
    L_M: float = .224
    n_p: int = 2


# %%
class InductionMotorVHzObsCtrl(Ctrl):
    """
    Observer-based V/Hz control for induction motors.

    Parameters
    ----------
    pars : InductionMotorObsVHzCtrlPars
        Control parameters.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        super().__init__()
        # Instantiate classes
        self.observer = SensorlessFluxObserver(pars)
        self.pwm = PWM(pars)
        self.rate_limiter = RateLimiter(pars)
        # Reference
        self.w_m_ref = pars.w_m_ref
        self.psi_s_ref = pars.psi_s_nom
        # Control Parameters
        self.T_s = pars.T_s
        self.alpha_f = pars.alpha_f
        self.alpha_r = pars.alpha_r
        self.alpha_psi = pars.alpha_psi
        self.n_p = pars.n_p
        self.k_tau = pars.k_tau
        self.i_s_max = pars.i_s_max
        self.slip_compensation = pars.slip_compensation
        # Motor parameters
        self.R_s = pars.R_s
        self.R_R = pars.R_R
        self.L_sgm = pars.L_sgm
        # Initial states
        self.theta_s, self.tau_M_ref, self.w_r_ref = 0, 0, 0

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : InductionMotorDrive
            Continuous-time model of an induction motor drive for getting the
            feedback signals.

        Returns
        -------
        T_s : float
            Sampling period.
        d_abc_ref : ndarray, shape (3,)
            Duty ratio references.

        """
        # Get the speed reference
        w_m_ref = self.rate_limiter(self.w_m_ref(self.t))

        # Measure the feedback signals
        i_s_abc = mdl.motor.meas_currents()  # Phase currents
        u_dc = mdl.conv.meas_dc_voltage()  # DC-bus voltage

        # Space vector and coordinate transformation
        i_s = np.exp(-1j*self.theta_s)*abc2complex(i_s_abc)

        # Get the states
        u_s = self.pwm.realized_voltage
        psi_R = self.observer.psi_R
        tau_M_ref = self.tau_M_ref
        w_r_ref = self.w_r_ref

        # Torque estimate (11c)
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_R))

        # Slip frequency compensation (if enabled) for the low-pass filter.
        # Note, could also be based on the low-pass filtered torque.
        psi_R_sqr = np.abs(psi_R)**2
        if self.slip_compensation and psi_R_sqr > 0:
            w_r = self.R_R*tau_M/(1.5*self.n_p*psi_R_sqr)
        else:
            w_r = 0

        # Slip compensation (9). Uses the low-pass filtered slip estimate
        # w_r_ref. Note if slip compensation disabled w_r_ref == 0.
        w_s_ref = w_m_ref + w_r_ref

        # Dynamic frequency (7a)
        w_s = w_s_ref - self.k_tau*(tau_M - tau_M_ref)

        # State feedback
        u_s_ref = self.state_feedback(i_s, psi_R, w_s)

        # Duty ratios
        d_abc_ref, u_s_ref_lim = self.pwm.output(
            u_s_ref, u_dc, self.theta_s, w_s)

        # Data logging
        data = Bunch(
            i_s=i_s,
            psi_s=psi_R + self.L_sgm*i_s,
            psi_s_ref=self.psi_s_ref,
            t=self.t,
            theta_s=self.theta_s,
            u_dc=u_dc,
            u_s=u_s,
            w_m=self.observer.w_m,
            w_m_ref=w_m_ref,
            w_s=w_s,
            tau_M=tau_M,
        )
        self.save(data)

        # Update the states
        self.pwm.update(u_s_ref_lim)
        self.observer.update(u_s, i_s, w_s)
        self.tau_M_ref += self.T_s*self.alpha_f*(tau_M - self.tau_M_ref)
        self.theta_s += self.T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.w_r_ref += self.T_s*self.alpha_r*(w_r - self.w_r_ref)
        self.update_clock(self.T_s)

        return self.T_s, d_abc_ref

    def state_feedback(self, i_s, psi_R, w_s):
        """Compute the stator voltage reference."""
        # Internal current reference for state feedback (6b)
        i_s_ref = (self.psi_s_ref - psi_R)/self.L_sgm
        # Limit the reference
        if np.abs(i_s_ref) > self.i_s_max:
            i_s_ref = self.i_s_max*i_s_ref/np.abs(i_s_ref)
        # State feedback (6a)
        u_s_ref = (
            self.R_s*i_s_ref + 1j*w_s*self.psi_s_ref +
            self.L_sgm*self.alpha_psi*(i_s_ref - i_s))
        return u_s_ref


# %%
class SensorlessFluxObserver:
    """
    Sensorless reduced-order flux observer.

    This observer is a variant of [1]_. The observer gain decouples the
    electrical and mechanical dynamics and allows placing the poles of the
    corresponding linearized estimation error dynamics. This implementation
    operates in controller coordinates.

    Parameters
    ----------
    pars : InductionMotorVHzObsCtrlPars
        Control parameters.

    References
    ----------
    .. [1] Hinkkanen, Harnefors, Luomi, "Reduced-order flux observers with
       stator-resistance adaptation for speed-sensorless induction motor
       drives," IEEE Trans. Power Electron., 2010,
       https://doi.org/10.1109/TPEL.2009.2039650

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        self.T_s = pars.T_s
        self.R_s = pars.R_s
        self.R_R = pars.R_R
        self.L_sgm = pars.L_sgm
        self.alpha = pars.R_R/pars.L_M
        self.alpha_o = pars.alpha_o
        self.zeta_inf = pars.zeta_inf
        # Initial states
        self.psi_R, self.i_s_old, self.w_m = 0j, 0j, 0

    def update(self, u_s, i_s, w_s):
        """
        Update the states of the observer.

        Parameters
        ----------
        u_s : complex
            Stator voltage.
        i_s : complex
            Stator current.
        w_s : float
            Angular frequency of the reference frame.

        """
        # Decay rate
        lambd = self.zeta_inf*np.abs(w_s) + .5*self.alpha
        # Observer gain (without the orthogonal projection which is
        # embedded into the state update)
        g_o = 2*lambd/(self.alpha - 1j*self.w_m)

        # Time derivative of the stator current
        di_s = (i_s - self.i_s_old)/self.T_s

        # Error voltage
        e = (
            self.L_sgm*(di_s + 1j*w_s*i_s) + (self.R_s + self.R_R)*i_s -
            (self.alpha - 1j*self.w_m)*self.psi_R - u_s)

        # Error signal
        err = e/self.psi_R if np.abs(self.psi_R) > 0 else 0

        # Update the states
        self.w_m -= self.T_s*self.alpha_o*err.imag
        self.psi_R += self.T_s*(
            u_s - self.R_s*i_s - self.L_sgm*di_s - 1j*w_s*
            (self.psi_R + self.L_sgm*i_s) + g_o*self.psi_R*err.real)
        self.i_s_old = i_s



"""
Observer-based V/Hz control for synchronous motor drives.

This method is based on [1]_.

References
----------
.. [1] Tiitinen, Hinkkanen, Kukkola, Routimo, Pellegrino, Harnefors, "Stable
    and passive observer-based V/Hz control for synchronous Motors," Proc.
    IEEE ECCE, Detroit, MI, Oct. 2022.

"""

from typing import Callable
from dataclasses import dataclass, field
import numpy as np
from motulator.helpers import abc2complex, Bunch
from motulator.control.common import Ctrl, PWM, RateLimiter
from motulator.control.sm_torque import TorqueCharacteristics


# %%
@dataclass
class SynchronousMotorVHzObsCtrlPars:
    """Control parameters."""

    # pylint: disable=too-many-instance-attributes
    # Speed reference (in electrical rad/s)
    w_m_ref: Callable[[float], float] = field(
        repr=False, default=lambda t: (t > .2)*(2*np.pi*75))

    # Control
    T_s: float = 250e-6
    psi_s_max: float = None
    psi_s_min: float = None
    rate_limit: float = np.inf
    i_s_max: float = 1.5*np.sqrt(2)*5
    alpha_psi: float = 2*np.pi*50
    alpha_tau: float = 2*np.pi*50
    alpha_f: float = 2*np.pi*1
    k_u: float = 1

    # Observer
    alpha_o: float = 2*np.pi*20
    zeta_inf: float = .2

    # Motor parameter estimates
    R_s: float = 3.6
    L_d: float = .036
    L_q: float = .051
    psi_f: float = .545
    n_p: int = 3


# %%
class SynchronousMotorVHzObsCtrl(Ctrl):
    """
    Observer-based V/Hz control for synchronous motors.

    Parameters
    ----------
    pars : SynchronousMotorVHzObsCtrlPars
        Control parameters.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        super().__init__()
        # Instantiate classes
        self.observer = SensorlessFluxObserver(pars)
        self.pwm = PWM(pars)
        self.rate_limiter = RateLimiter(pars)
        self.flux_torque_ref = FluxTorqueRef(pars)
        # Reference
        self.w_m_ref = pars.w_m_ref
        # Motor parameters
        self.R_s = pars.R_s
        self.n_p = pars.n_p
        # Controller parameters
        self.T_s = pars.T_s
        self.alpha_f = pars.alpha_f
        self.alpha_psi = pars.alpha_psi
        # Gain k_tau
        G = (pars.L_d - pars.L_q)/(pars.L_d*pars.L_q)
        psi_s0 = pars.psi_f if pars.psi_f > 0 else pars.psi_s_min
        if pars.psi_f > 0:  # PMSM or PM-SyRM
            c_delta0 = 1.5*pars.n_p*(pars.psi_f*psi_s0/pars.L_d - G*psi_s0**2)
        else:  # SyRM
            c_delta0 = 1.5*pars.n_p*G*psi_s0**2
        self.k_tau = pars.alpha_tau/c_delta0
        # Initial states
        self.theta_s, self.tau_M_ref = 0, 0

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : SynchronousMotorDrive
            Continuous-time model of a synchronous motor drive for getting the
            feedback signals.

        Returns
        -------
        T_s : float
            Sampling period.
        d_abc_ref : ndarray, shape (3,)
            Duty ratio references.

        """
        # Get the speed reference
        w_m_ref = self.rate_limiter(self.w_m_ref(self.t))

        # Measure the feedback signals
        i_s_abc = mdl.motor.meas_currents()  # Phase currents
        u_dc = mdl.conv.meas_dc_voltage()  # DC-bus voltage

        # Space vector and coordinate transformation
        i_s = np.exp(-1j*self.theta_s)*abc2complex(i_s_abc)

        # Get the states
        u_s = self.pwm.realized_voltage
        psi_s = self.observer.psi_s
        tau_M_ref = self.tau_M_ref

        # Limited flux and torque references
        psi_s_ref, _ = self.flux_torque_ref(tau_M_ref, w_m_ref, u_dc)

        # Electromagnetic torque (7d)
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_s))

        # Dynamic frequency (5a)
        w_s = w_m_ref - self.k_tau*(tau_M - tau_M_ref)

        # Voltage reference (4)
        err = psi_s_ref - psi_s
        u_s_ref = self.R_s*i_s + 1j*w_s*psi_s_ref + self.alpha_psi*err

        # Duty ratios
        d_abc_ref, u_s_ref_lim = self.pwm.output(
            u_s_ref, u_dc, self.theta_s, w_s)

        # Data logging
        data = Bunch(
            i_s=i_s,
            psi_s=psi_s,
            psi_s_ref=psi_s_ref,
            t=self.t,
            theta_s=self.theta_s,
            u_dc=u_dc,
            u_s=u_s,
            w_m_ref=w_m_ref,
            w_s=w_s,
            tau_M=tau_M,
        )
        self.save(data)

        # Update the states
        self.pwm.update(u_s_ref_lim)
        self.observer.update(u_s, i_s, w_s)
        self.tau_M_ref += self.T_s*self.alpha_f*(tau_M - self.tau_M_ref)
        self.theta_s += self.T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.update_clock(self.T_s)

        return self.T_s, d_abc_ref


# %%
class FluxTorqueRef:
    """
    Flux and torque references.

    Parameters
    ----------
    pars : SynchronousMotorVHzObsCtrlPars
        Control parameters.

    """

    def __init__(self, pars):
        self.psi_s_min = (
            pars.psi_f if pars.psi_s_min is None else pars.psi_s_min)
        self.psi_s_max = np.inf if pars.psi_s_max is None else pars.psi_s_max
        self.k_u = pars.k_u
        # Merged MTPV and current limits
        tq = TorqueCharacteristics(pars)
        lims = tq.mtpv_and_current_limits(i_s_max=pars.i_s_max)
        self.tau_M_lim = lims.tau_M_vs_abs_psi_s
        # MTPA locus
        mtpa = tq.mtpa_locus(i_s_max=pars.i_s_max)
        self.psi_s_mtpa = mtpa.abs_psi_s_vs_tau_M

    def __call__(self, tau_M_ref, w_m, u_dc):
        """
        Calculate the stator flux reference and limit the torque reference.

        Parameters
        ----------
        tau_M_ref : float
            Unlimited torque reference.
        w_m : float
            Rotor speed or its reference (in electrical rad/s).
        u_dc : float
            DC-bus voltage.

        Returns
        -------
        psi_s_ref : float
            Stator flux reference.
        tau_M_ref_lim : float
            Limited torque reference.

        """
        # Get the MTPA flux
        psi_s_mtpa = self.psi_s_mtpa(np.abs(tau_M_ref))
        np.clip(psi_s_mtpa, self.psi_s_min, self.psi_s_max, out=psi_s_mtpa)

        # Field weakening
        u_s_max = self.k_u*u_dc/np.sqrt(3)
        psi_s_max = u_s_max/np.abs(w_m) if np.abs(w_m) > 0 else np.inf

        # Flux reference
        psi_s_ref = np.min([psi_s_max, psi_s_mtpa])

        # Limit the torque reference according to the MTPV and current limits
        tau_M_lim = self.tau_M_lim(psi_s_ref)
        tau_M_ref_lim = np.min([tau_M_lim, np.abs(tau_M_ref)
                                ])*np.sign(tau_M_ref)

        return psi_s_ref, tau_M_ref_lim


# %%
class SensorlessFluxObserver:
    """
    Sensorless stator flux observer.

    This observer is a variant of [1]_. The observer gain decouples the
    electrical and mechanical dynamics and allows placing the poles of the
    corresponding linearized estimation error dynamics. For simplicity, the
    current model is here implemented in rotor coordinates, however this is
    mathematically equivalent to controller coordinates implementation in [2]_.

    Parameters
    ----------
    pars : SynchronousMotorObsVHzCtrlPars
        Control parameters.

    References
    ----------
    .. [1] Hinkkanen, Saarakkala, Awan, Mölsä, Tuovinen, "Observers for
        sensorless synchronous motor drives: Framework for design and
        analysis," IEEE Trans. Ind. Appl., 2018,
        https://doi.org/10.1109/TIA.2018.2858753
    .. [2] Tiitinen, Hinkkanen, Kukkola, Routimo, Pellegrino, Harnefors,
        "Stable and passive observer-based V/Hz control for synchronous
        Motors," Proc. IEEE ECCE, Detroit, MI, Oct. 2022.

    """

    def __init__(self, pars):
        self.T_s = pars.T_s
        self.R_s = pars.R_s
        self.L_d = pars.L_d
        self.L_q = pars.L_q
        self.psi_f = pars.psi_f
        self.alpha_o = pars.alpha_o
        self.b_p = .5*pars.R_s*(pars.L_d + pars.L_q)/(pars.L_d*pars.L_q)
        self.zeta_inf = pars.zeta_inf
        # Initial states
        self.delta, self.psi_s = 0, pars.psi_f

    def update(self, u_s, i_s, w_s):
        """
        Update the states for the next sampling period.

        Parameters
        ----------
        u_s : complex
            Stator voltage.
        i_s : complex
            Stator current.
        w_s : float
            Stator angular frequency.

        """
        # Transformations to rotor coordinates
        i_sr = i_s*np.exp(1j*self.delta)
        psi_sr = self.psi_s*np.exp(1j*self.delta)

        # Auxiliary flux and estimation error in rotor coordinates
        psi_ar = self.psi_f + (self.L_d - self.L_q)*np.conj(i_sr)
        e_r = self.L_d*i_sr.real + 1j*self.L_q*i_sr.imag + self.psi_f - psi_sr

        # Auxiliary flux in controller coordinates
        psi_a = np.exp(-1j*self.delta)*psi_ar

        g_o = self.b_p + 2*self.zeta_inf*np.abs(w_s)

        if np.abs(psi_ar) > 0:
            # Correction voltage in controller coordinates
            v = g_o*psi_a*np.real(e_r/psi_ar)
            # Error signal
            w_delta = self.alpha_o*np.imag(e_r/psi_ar)
        else:
            v, w_delta = 0, 0

        # Update the states
        self.psi_s += self.T_s*(u_s - self.R_s*i_s - 1j*w_s*self.psi_s + v)
        self.delta += self.T_s*w_delta



"""
Flux-vector control for synchronous motor drives.

This implements a version of stator-flux-vector control [1]_. Rotor coordinates 
as well as decoupling between the stator flux and torque channels are used [2]_. 
Here, the stator flux magnitude and the electromagnetic torque are selected as
controllable variables. Proportional controllers are used for simplicity. The 
magnetic saturation is not considered in this implementation.

References
----------
.. [1] Pellegrino, Armando, Guglielmi, “Direct flux field-oriented control of
   IPM drives with variable DC link in the field-weakening region,” IEEE Trans.
   Ind. Appl., 2009, https://doi.org/10.1109/TIA.2009.2027167

.. [2] Awan, Hinkkanen, Bojoi, Pellegrino, "Stator-flux-oriented control of
   synchronous motors: A systematic design procedure," IEEE Trans. Ind. Appl.,
   2019, https://doi.org/10.1109/TIA.2019.2927316

"""
from typing import Callable
from dataclasses import dataclass, field
import numpy as np
from motulator.helpers import abc2complex, Bunch
from motulator.control.common import Ctrl, SpeedCtrl, PWM
from motulator.control.sm_vector import SensorlessObserver
from motulator.control.sm_obs_vhz import FluxTorqueRef


# %%
@dataclass
class SynchronousMotorFluxVectorCtrlPars:
    """Control parameters: flux-vector control for synchronous motor drives."""

    # pylint: disable=too-many-instance-attributes
    # Speed reference (in electrical rad/s)
    w_m_ref: Callable[[float], float] = field(
        repr=False, default=lambda t: (t > .2)*(2*np.pi*75))
    # Mode
    sensorless: bool = True
    # Sampling period
    T_s: float = 250e-6
    # Flux reference limits
    psi_s_min: float = None
    psi_s_max: float = None
    # Voltage marginal
    k_u: float = .9
    # Bandwidths
    alpha_psi: float = 2*np.pi*150
    alpha_tau: float = 2*np.pi*50
    alpha_s: float = 2*np.pi*4
    # Maximum values
    tau_M_max: float = 1.5*14
    i_s_max: float = 1.5*np.sqrt(2)*5.
    # Motor parameter estimates
    R_s: float = 3.6
    L_d: float = .036
    L_q: float = .051
    psi_f: float = .545
    n_p: int = 3
    J: float = .015
    # Sensorless observer (used only in the sensorless mode)
    w_o: float = 2*np.pi*100
    zeta_inf: float = .2
    # Sensored observer (used only in the sensored mode)
    g: float = 2*np.pi*15


# %%
class SynchronousMotorFluxVectorCtrl(Ctrl):
    """
    Flux-vector control for synchronous motor drives.

    This class interconnects the subsystems of the control system and
    provides the interface to the solver.

    Parameters
    ----------
    pars : SynchronousMotoroFluxVectorCtrlPars
        Control parameters.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        super().__init__()
        self.T_s = pars.T_s
        self.w_m_ref = pars.w_m_ref
        self.sensorless = pars.sensorless
        self.speed_ctrl = SpeedCtrl(pars)
        self.pwm = PWM(pars)
        if pars.sensorless:
            self.observer = SensorlessObserver(pars)
        else:
            self.observer = Observer(pars)
        self.flux_torque_ref = FluxTorqueRef(pars)
        # Bandwidths
        self.alpha_psi = pars.alpha_psi
        self.alpha_tau = pars.alpha_tau
        # Motor parameter estimates
        self.R_s = pars.R_s
        self.L_d = pars.L_d
        self.L_q = pars.L_q
        self.n_p = pars.n_p

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : SynchronousMotorDrive
            Continuous-time model of a synchronous motor drive for getting the
            feedback signals.

        Returns
        -------
        T_s : float
            Sampling period.
        d_abc_ref : ndarray, shape (3,)
            Duty ratio references.

        """
        # Get the speed reference
        w_m_ref = self.w_m_ref(self.t)

        # Feedback signals
        i_s_abc = mdl.motor.meas_currents()  # Phase currents
        u_dc = mdl.conv.meas_dc_voltage()  # DC-bus voltage
        u_s = self.pwm.realized_voltage  # Realized voltage from PWM

        if self.sensorless:
            # Get the rotor speed and position estimates
            w_m, theta_m = self.observer.w_m, self.observer.theta_m
        else:
            # Measure the rotor speed
            w_m = self.n_p*mdl.mech.meas_speed()
            # Limit the electrical rotor position into [-pi, pi)
            theta_m = np.mod(
                self.n_p*mdl.mech.meas_position() + np.pi, 2*np.pi) - np.pi

        # Current vector in estimated rotor coordinates
        i_s = np.exp(-1j*theta_m)*abc2complex(i_s_abc)

        # Flux and torque estimates
        psi_s = self.observer.psi_s
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_s))

        # Outputs
        tau_M_ref = self.speed_ctrl.output(w_m_ref/self.n_p, w_m/self.n_p)
        psi_s_ref, tau_M_ref_lim = self.flux_torque_ref(tau_M_ref, w_m, u_dc)

        # Auxiliary current
        i_a = psi_s.real/self.L_q + 1j*psi_s.imag/self.L_d - i_s

        # Torque-production factor (c_tau = 0 corresponds to the MTPV condition)
        c_tau = np.real(i_a*np.conj(psi_s))

        # References for the flux and torque controllers
        v_psi = self.alpha_psi*(psi_s_ref - np.abs(psi_s))
        v_tau = self.alpha_tau*(tau_M_ref_lim - tau_M)
        if c_tau > 0:
            v = (np.abs(psi_s)*i_a*v_psi + 1j*psi_s*v_tau)/c_tau
        else:
            v = v_psi

        # Stator voltage reference
        u_s_ref = self.R_s*i_s + 1j*w_m*psi_s + v

        # PWM output
        d_abc_ref, u_s_ref_lim = self.pwm.output(u_s_ref, u_dc, theta_m, w_m)

        # Data logging
        data = Bunch(
            i_s=i_s,
            psi_s=psi_s,
            psi_s_ref=psi_s_ref,
            t=self.t,
            tau_M_ref_lim=tau_M_ref_lim,
            theta_m=theta_m,
            u_dc=u_dc,
            u_s=u_s,
            w_m=w_m,
            w_m_ref=w_m_ref,
        )
        self.save(data)

        # Update states
        self.observer.update(u_s, i_s, w_m)
        self.speed_ctrl.update(tau_M_ref_lim)
        self.pwm.update(u_s_ref_lim)
        self.update_clock(self.T_s)

        return self.T_s, d_abc_ref


# %%
class Observer:
    """
    Sensored observer.

    Parameters
    ----------
    pars : SynchronousMotoroFluxVectorCtrlPars
        Control parameters.

    """

    def __init__(self, pars):
        self.T_s = pars.T_s
        self.R_s = pars.R_s
        self.L_d = pars.L_d
        self.L_q = pars.L_q
        self.psi_f = pars.psi_f
        self.g = pars.g
        # Initial state
        self.psi_s = pars.psi_f

    def update(self, u_s, i_s, w_m):
        """
        Update the states for the next sampling period.

        Parameters
        ----------
        u_s : complex
            Stator voltage in estimated rotor coordinates.
        i_s : complex
            Stator current in estimated rotor coordinates.
        w_m : float
            Rotor speed (in electrical rad/s).

        """
        # Estimation error
        e = self.L_d*i_s.real + 1j*self.L_q*i_s.imag + self.psi_f - self.psi_s

        # Update the state
        self.psi_s += self.T_s*(
            u_s - self.R_s*i_s - 1j*w_m*self.psi_s + self.g*e)



"""Import and plot flux maps from the SyR-e project."""

# %%
import numpy as np
import matplotlib.pyplot as plt
from cycler import cycler
from scipy.io import loadmat
from scipy.interpolate import griddata
from motulator.helpers import Bunch

# Plotting parameters
plt.rcParams['axes.prop_cycle'] = cycler(color='brgcmyk')
plt.rcParams['lines.linewidth'] = 1.
plt.rcParams['axes.grid'] = True
plt.rcParams.update({'text.usetex': False})


# %%
def import_syre_data(fname='THOR.mat', add_negative_q_axis=True):
    """
    Import a flux map from the MATLAB data file in the SyR-e format.

    For more information on the SyR-e project and the MATLAB file format,
    please visit:

        https://github.com/SyR-e/syre_public

    The imported data is converted to the PMSM coordinate convention, in which
    the PM flux is along the d axis.

    Parameters
    ----------
    fname : str, optional
        MATLAB file name. The default is 'THOR.mat'.
    add_negative_q_axis : bool, optional
        Adds the negative q-axis data based on the symmetry.

    Returns
    -------
    Bunch object with the following fields defined:
    i_s : complex ndarray
        Stator current data.
    psi_s : complex ndarray
        Stator flux linkage data.
    tau_M : ndarray

    Notes
    -----
    Some example data files (including THOR.mat) are available in the SyR-e
    repository, licensed under the Apache License, Version 2.0.

    """
    # Read the data from mat-file
    mat = loadmat(fname)

    # Use the PMSM convention in coordinates. Rotate the matrices by -90 deg
    # to get monotonically increasing grid data.
    i_d = -np.rot90(mat['motorModel'][0, 0]['FluxMap_dq'][0, 0]['Iq'], -1)
    i_q = np.rot90(mat['motorModel'][0, 0]['FluxMap_dq'][0, 0]['Id'], -1)
    psi_d = -np.rot90(mat['motorModel'][0, 0]['FluxMap_dq'][0, 0]['Fq'], -1)
    psi_q = np.rot90(mat['motorModel'][0, 0]['FluxMap_dq'][0, 0]['Fd'], -1)
    tau_M = np.rot90(mat['motorModel'][0, 0]['FluxMap_dq'][0, 0]['T'], -1)

    # Clip the negative q-axis values
    np.clip(i_q, 0, np.inf, out=i_q)
    np.clip(psi_q, 0, np.inf, out=psi_q)

    if add_negative_q_axis:
        # Add the negative q-axis data, flipped based on the symmetry
        i_d = np.concatenate((np.flipud(i_d), i_d))
        i_q = np.concatenate((-np.flipud(i_q), i_q))
        psi_d = np.concatenate((np.flipud(psi_d), psi_d))
        psi_q = np.concatenate((-np.flipud(psi_q), psi_q))
        tau_M = np.concatenate((-np.flipud(tau_M), tau_M))

    # Pack the data in the complex form
    i_s = i_d + 1j*i_q
    psi_s = psi_d + 1j*psi_q

    return Bunch(i_s=i_s, psi_s=psi_s, tau_M=tau_M)


# %%
def plot_flux_map(data):
    """
    Plot the flux linkage as function of the current.

    Parameters
    ----------
    data : Bunch
        Flux map data.

    """
    fig = plt.figure(figsize=(10, 5))
    ax1 = fig.add_subplot(1, 2, 1, projection='3d')
    ax2 = fig.add_subplot(1, 2, 2, projection='3d')
    ax1.plot_surface(data.i_s.real, data.i_s.imag, data.psi_s.real)
    ax2.plot_surface(data.i_s.real, data.i_s.imag, data.psi_s.imag)
    ax1.set_xlabel(r'$i_\mathrm{d}$ (A)')
    ax1.set_ylabel(r'$i_\mathrm{q}$ (A)')
    ax1.set_zlabel(r'$\psi_\mathrm{d}$ (Vs)')
    ax2.set_xlabel(r'$i_\mathrm{d}$ (A)')
    ax2.set_ylabel(r'$i_\mathrm{q}$ (A)')
    ax2.set_zlabel(r'$\psi_\mathrm{q}$ (Vs)')


# %%
def plot_torque_map(data):
    """
    Plot the torque as function of the current.

    Parameters
    ----------
    data : Bunch
        Flux map data.

    """
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.plot_surface(data.i_s.real, data.i_s.imag, data.tau_M)
    ax.set_xlabel(r'$i_\mathrm{d}$ (A)')
    ax.set_ylabel(r'$i_\mathrm{q}$ (A)')
    ax.set_zlabel(r'$\tau_\mathrm{M}$ (Nm)')


# %%
def plot_flux_vs_current(data):
    """
    Plot the flux vs. current characteristics.

    Parameters
    ----------
    data : Bunch
        Flux map data.

    """
    # Indices corresponding to i_d = 0 and i_q = 0
    ind_d_0 = np.argmin(np.abs(data.i_s.real[0, :]))
    ind_q_0 = np.argmin(np.abs(data.i_s.imag[:, 0]))
    # Indices corresponding to min(i_d) and max(i_d)
    ind_d_min = np.argmin(data.i_s.real[0, :])
    ind_d_max = np.argmax(data.i_s.real[0, :])

    fig = plt.figure()
    ax = fig.add_subplot()
    ax.plot(
        data.i_s.real[ind_q_0, :],
        data.psi_s.real[ind_q_0, :],
        color='r',
        linestyle='-',
        label=r'$\psi_\mathrm{d}(i_\mathrm{d}, 0)$')
    ax.plot(
        data.i_s.real[-1, :],
        data.psi_s.real[-1, :],
        color='r',
        linestyle='--',
        label=r'$\psi_\mathrm{d}(i_\mathrm{d}, i_\mathrm{q,max})$')
    ax.plot(
        data.i_s.imag[:, ind_d_0],
        data.psi_s.imag[:, ind_d_0],
        color='b',
        linestyle='-',
        label=r'$\psi_\mathrm{q}(0, i_\mathrm{q})$')
    ax.plot(
        data.i_s.imag[:, ind_d_min],
        data.psi_s.imag[:, ind_d_min],
        color='b',
        linestyle=':',
        label=r'$\psi_\mathrm{q}(i_\mathrm{d,min}, i_\mathrm{q})$')
    ax.plot(
        data.i_s.imag[:, ind_d_max],
        data.psi_s.imag[:, ind_d_max],
        color='b',
        linestyle='--',
        label=r'$\psi_\mathrm{q}(i_\mathrm{d,max}, i_\mathrm{q})$')

    ax.set_xlabel(r'$i_\mathrm{d}$, $i_\mathrm{q}$ (A)')
    ax.set_ylabel(r'$\psi_\mathrm{d}$, $\psi_\mathrm{q}$ (Vs)')
    ax.legend()


# %%
def downsample_flux_map(data, N_d=32, N_q=32):
    """
    Downsample the flux map by means of linear interpolation.

    Parameters
    ----------
    data : Bunch
        Flux map data.
    N_d : int, optional
        Number of interpolated samples in the d axis. The default is 32.
    N_q : int, optional
        Number of interpolated samples in the q axis. The default is 32.

    Returns
    -------
    Bunch object with the following fields defined:
    i_s : complex ndarray, shape (N_d, N_q)
        Stator current data.
    psi_s : complex ndarray, shape (N_d, N_q)
        Stator flux linkage data.
    tau_M : ndarray, shape (N_d, N_q)

    """
    # Points at which to interpolate data
    i_d = np.linspace(np.min(data.i_s.real), np.max(data.i_s.real), N_d)
    i_q = np.linspace(np.min(data.i_s.imag), np.max(data.i_s.imag), N_q)
    i_d, i_q = np.meshgrid(i_d, i_q)

    # Interpolate data
    points = (np.ravel(data.i_s.real), np.ravel(data.i_s.imag))
    psi_s = griddata(points, np.ravel(data.psi_s), (i_d, i_q), method='linear')
    tau_M = griddata(points, np.ravel(data.tau_M), (i_d, i_q), method='linear')
    i_s = i_d + 1j*i_q

    return Bunch(i_s=i_s, psi_s=psi_s, tau_M=tau_M)


# %%
def invert_flux_map(data, N_d=32, N_q=32):
    """
    Compute the inverse flux map by means of linear interpolation.

    Parameters
    ----------
    data : Bunch
        Flux map data.
    N_d : int, optional
        Number of interpolated samples in the d axis. The default is 32.
    N_q : int, optional
        Number of interpolated samples in the q axis. The default is 32.

    Returns
    -------
    Bunch object with the following fields defined:
    psi_s : complex ndarray, shape (N_d, N_q)
        Stator flux linkage data.
    i_s : complex ndarray, shape (N_d, N_q)
        Stator current data.
    tau_M : ndarray, shape (N_d, N_q)

    """
    # Get the range
    psi_d_max = np.min(np.max(data.psi_s.real, axis=0))
    psi_d_min = np.max(np.min(data.psi_s.real, axis=0))
    psi_q_max = np.min(np.max(data.psi_s.imag, axis=1))
    psi_q_min = np.max(np.min(data.psi_s.imag, axis=1))

    # Points at which to interpolate data
    psi_d = np.linspace(psi_d_min, psi_d_max, N_d)
    psi_q = np.linspace(psi_q_min, psi_q_max, N_q)
    psi_d, psi_q = np.meshgrid(psi_d, psi_q)

    # Interpolate data
    points = (np.ravel(data.psi_s.real), np.ravel(data.psi_s.imag))
    i_s = griddata(points, np.ravel(data.i_s), (psi_d, psi_q), method='linear')
    tau_M = griddata(
        points, np.ravel(data.tau_M), (psi_d, psi_q), method='linear')
    psi_s = psi_d + 1j*psi_q

    return Bunch(psi_s=psi_s, i_s=i_s, tau_M=tau_M)



"""
Sensorless control with signal injection for synchronous motor drives.

This module contains a simple example of square-wave signal injection for low-
speed operation. A phase-locked loop is used to track the rotor position. For
a wider speed range, signal injection could be combined to a model-based
observer. The effects of magnetic saturation are not compensated for in this
version.

"""

from dataclasses import dataclass
import numpy as np
from motulator.helpers import abc2complex, Bunch
from motulator.control.common import Ctrl, SpeedCtrl, PWM
from motulator.control.sm_vector import (
    CurrentCtrl, CurrentRef, SynchronousMotorVectorCtrlPars)


# %%
@dataclass
class SynchronousMotorSignalInjectionCtrlPars(SynchronousMotorVectorCtrlPars):
    """Square-wave signal injection parameters."""

    U_inj: float = 200


# %%
class SynchronousMotorSignalInjectionCtrl(Ctrl):
    """
    Sensorless control with signal injection for a synchronous motor drive.

    This class interconnects the subsystems of the control system and
    provides the interface to the solver.

    Parameters
    ----------
    pars : SynchronousMotorVectorCtrlPars
        Control parameters.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pars):
        super().__init__()
        self.T_s = pars.T_s
        self.w_m_ref = pars.w_m_ref
        self.n_p = pars.n_p
        self.current_ctrl = CurrentCtrl(pars)
        self.speed_ctrl = SpeedCtrl(pars)
        self.current_ref = CurrentRef(pars)
        self.pwm = PWM(pars)
        self.pll = PhaseLockedLoop(pars)
        self.signal_inj = SignalInjection(pars)

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : SynchronousMotorDrive
            Continuous-time model of a synchronous motor drive for getting the
            feedback signals.

        Returns
        -------
        T_s : float
            Sampling period.
        d_abc_ref : ndarray, shape (3,)
            Duty ratio references.

        """
        # Get the speed reference
        w_m_ref = self.w_m_ref(self.t)

        # Measure the feedback signals
        i_s_abc = mdl.motor.meas_currents()  # Phase currents
        u_dc = mdl.conv.meas_dc_voltage()  # DC-bus voltage

        # Get the rotor speed and position estimates
        w_m, theta_m = self.pll.w_m, self.pll.theta_m

        # Current vector in estimated rotor coordinates
        i_s = np.exp(-1j*theta_m)*abc2complex(i_s_abc)

        # Filter the current measurement for the current controller
        i_s_filt = self.signal_inj.filter_current(i_s)

        # Outputs
        tau_M_ref = self.speed_ctrl.output(w_m_ref/self.n_p, w_m/self.n_p)
        i_s_ref, tau_M_ref_lim = self.current_ref.output(tau_M_ref, w_m, u_dc)
        err = self.signal_inj.output(i_s.imag)
        # Superimpose the excitation voltage on the d-axis
        u_s_ref = self.current_ctrl.output(
            i_s_ref, i_s_filt) + self.signal_inj.u_sd_inj
        d_abc_ref, u_s_ref_lim = self.pwm.output(u_s_ref, u_dc, theta_m, w_m)

        # Data logging
        data = Bunch(
            i_s=i_s_filt,
            i_s_ref=i_s_ref,
            t=self.t,
            tau_M_ref_lim=tau_M_ref_lim,
            theta_m=theta_m,
            u_dc=u_dc,
            u_s=u_s_ref_lim,
            w_m=w_m,
            w_m_ref=w_m_ref,
        )
        self.save(data)

        # Update states
        self.speed_ctrl.update(tau_M_ref_lim)
        self.current_ref.update(tau_M_ref_lim, u_s_ref, u_dc)
        self.current_ctrl.update(u_s_ref_lim, w_m)
        self.pwm.update(u_s_ref_lim)
        self.signal_inj.update(i_s)
        self.pll.update(err)
        self.update_clock(self.T_s)

        return self.T_s, d_abc_ref


# %%
class SignalInjection:
    """
    Estimate the rotor position error based on signal injection.

    This signal injection method estimates the rotor position error based on
    the injected switching frequency signal, according to [1]_. The estimate
    can be used in a phase-locked loop or in a state observer to robustify
    low-speed sensorless operation.

    Parameters
    ----------
    pars : SynchronousMotorSignalInjectionCtrlPars
        Control parameters.

    References
    ----------
    .. [1] Kim, Ha, Sul, "PWM switching frequency signal injection sensorless
       method in IPMSM," IEEE Trans. Ind. Appl., 2012,
       https://doi.org/10.1109/TIA.2012.2210175

    """

    def __init__(self, pars):
        self.u_sd_inj = pars.U_inj
        # Initial states
        self.i_s_old, self.i_s_older = 0, 0
        # Error gain
        self.k = .5*pars.L_d*pars.L_q/((pars.L_q - pars.L_d)*pars.T_s)

    def output(self, i_sq):
        """
        Compute the rotor position estimation error.

        Parameters
        ----------
        i_sq : float
            Stator current q-component in estimated rotor coordinates.

        Returns
        -------
        err : float
            Rotor position estimation error.

        """
        di_sq = i_sq - 2*self.i_s_old.imag + self.i_s_older.imag
        err = self.k*di_sq/self.u_sd_inj if np.abs(self.u_sd_inj) > 0 else 0
        return err

    def update(self, i_s):
        """
        Store the old current values for the next sampling period.

        Parameters
        ----------
        i_s : complex
            Stator current in estimated rotor coordinates.

        """
        # Update the integral states
        self.i_s_older = self.i_s_old
        self.i_s_old = i_s
        # Reverse the d-axis square wave injection voltage
        self.u_sd_inj = -self.u_sd_inj

    def filter_current(self, i_s):
        """
        Filter the stator current using the previously measured value.

        Parameters
        ----------
        i_s : complex
            Unfiltered stator current in estimated rotor coordinates.

        Returns
        -------
        i_s_filt : complex
            Filtered stator current in estimated rotor coordinates.

        """
        # Filter currents
        i_s_filt = .5*(i_s + self.i_s_old)
        return i_s_filt


# %%
class PhaseLockedLoop:
    """
    Simple phase-locked loop for rotor-position estimation.

    Parameters
    ----------
    pars : SynchronousMotorVectorCtrlPars
        Control parameters.

    """

    def __init__(self, pars):
        self.T_s = pars.T_s
        # Gains
        self.k_p = 2*pars.w_o
        self.k_i = pars.w_o**2
        # # Initial states
        self.theta_m, self.w_m = 0, 0

    def update(self, err):
        """
        Update the states for the next sampling period.

        Parameters
        ----------
        err : Rotor position error.

        """
        # Speed estimation
        w_m = self.k_p*err + self.w_m

        # Update the states
        self.w_m += self.T_s*self.k_i*err
        self.theta_m += self.T_s*w_m  # Next line: limit into [-pi, pi)
        self.theta_m = np.mod(self.theta_m + np.pi, 2*np.pi) - np.pi
